/*++

	Copyright (c) 2009-2010 SMSC. All rights reserved.

	The information contained in this file is confidential and proprietary to
	Standard Microsystems Corporation.	No part of this file may be reproduced
	or distributed, in any form or by any means for any purpose, without the
	express written permission of Standard Microsystems Corporation.

	
Module Name:

	interrupt.c
		
Abstract:

	This module contains miniport functions for handling interrupt
	transactions and other helper routines called by these miniport 
	functions.
	
Revision History:
	01/19/2010	Initial version.

Notes:

--*/


#include "smsc7500.h"

#ifdef USE_INT_PIPE

/*++

Routine Description:

   Workitem to post all the free int requests to the target
   driver. This workitem is scheduled from the MiniportInitialize
   worker routine during start and from the NICFreeICB routine
   whenever the outstanding int requests to the target driver goes 
   below the watermark.
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - Pointer to our Adapter

Return Value:

	None.

--*/
VOID
NICPostIntsWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)

{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	NTSTATUS		ntStatus;
	PICB			pICB = NULL;

	SMSC_TRACE0(DBG_INT,"--> NICPostIntsWorkItemCallBack\r\n"); 
	
	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one. Clear it before processing to avoid 
	// race condition which can leave the device in a state where
	// no further int transactions are submitted
	//
	InterlockedExchange(&Adapter->IsPostIntsWorkItemQueued, FALSE);

	// Now lets actually do the interrupt submissions
	NdisAcquireSpinLock(&Adapter->IntLock);

	while(MP_IS_READY(Adapter) && !IsListEmpty(&Adapter->IntFreeList) && MP_TEST_FLAG(Adapter, fMP_POST_INTS))
	{
		pICB = (PICB) RemoveHeadList(&Adapter->IntFreeList);

		ASSERT(pICB);// cannot be NULL
		
		//
		// Insert the ICB in the Int busy queue
		//	   
		InsertTailList(&Adapter->IntBusyList, &pICB->List);
		NdisInterlockedIncrement(&Adapter->nBusyInt);		   
		ASSERT(Adapter->nBusyInt <= Adapter->MaxICBs);
		
		NdisReleaseSpinLock(&Adapter->IntLock);

		Adapter->nIntsPosted++;
		ntStatus = NICPostIntRequest(Adapter, pICB);
		
		if (!NT_SUCCESS ( ntStatus ) ) 
		{
			SMSC_TRACE1(DBG_ERROR, "NICPostIntRequest failed %x\r\n", ntStatus);

			// NICFreeICB expects it in the Processed queue, so put it there	
			// before making the call.
			NdisAcquireSpinLock(&Adapter->IntLock);

			// Remove it from the busy list 
			RemoveEntryList(&pICB->List);	
			NdisInterlockedDecrement(&Adapter->nBusyInt);
			ASSERT(Adapter->nBusyInt >= 0);

			// Put it in the Processed list.
			InsertTailList(&Adapter->IntProcessedList, &pICB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedInt);
			ASSERT(Adapter->nProcessedInt <= Adapter->MaxICBs);

			SMSC_TRACE1(DBG_INT,"ICB %p removed from the busy and inserted into IntProcessedList\r\n", pICB);

			NdisReleaseSpinLock(&Adapter->IntLock); 

			if (NdisInterlockedDecrement(&pICB->Ref) == 0) 
			{
				NICFreeICB(pICB);
			}
			else
			{
				// This should never happen
				SMSC_TRACE0(DBG_ERROR, "NICPostIntsWorkItemCallBack finds ICB with invalid Ref count!\r\n");			  
			}

			NdisAcquireSpinLock(&Adapter->IntLock);
			break;			  
		}
		
		NdisAcquireSpinLock(&Adapter->IntLock);
		
	}

	NdisReleaseSpinLock(&Adapter->IntLock);

	MP_DEC_REF(Adapter);	 

	SMSC_TRACE0(DBG_INT,"<-- NICPostIntsWorkItemCallBack\r\n"); 	   
}


/*++

Routine Description:

	This function submits all interrupt requests.
	
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	
Return value:

	NIC_STATUS_FAILURE if reset pipe failed, NIC_STATUS_SUCCESS otherwise
--*/

NIC_STATUS
NICStartInterruptPipe(
	PMP_ADAPTER Adapter
	)
{ 
	NIC_STATUS		nStatus = NIC_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_INT,"--> NICStartInterruptPipe\r\n");
	

	// We need to reset the pipe to make sure the data toggles are in sync.
	if(!NICResetPipe(Adapter, Adapter->Interrupt)) 
	{
		SMSC_TRACE0(DBG_ERROR, "Interrupt Pipe Reset failed !!! Aborting Int Pipe Start sequence.\r\n");
		nStatus = NIC_STATUS_FAILURE;
		goto Exit_NICStartInterruptPipe;
	}

	Adapter->Interrupt.PipeTransferAborted=0;
	NdisSetEvent(&Adapter->IsIntsCompletionWorkItemReady);

	MP_SET_FLAG(Adapter, fMP_POST_INTS);
	Adapter->IsIntReady=TRUE;

	NICPostInts(Adapter);

Exit_NICStartInterruptPipe:
	SMSC_TRACE0(DBG_INT,"<-- NICStartInterruptPipe\r\n");

	return (nStatus);
}


/*++
Routine Description:

   Post all the free int requests on USB if the flag to do so is enabled.
	  
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	
Return Value:

	None.

--*/
void
NICPostInts(
	PMP_ADAPTER Adapter
	)

{
	
	SMSC_TRACE0(DBG_INT,"--> NICPostInts\r\n");

	// If a workitem already scheduled don't queue another.		
	if (InterlockedExchange(&Adapter->IsPostIntsWorkItemQueued, TRUE) == FALSE)
	{
		Adapter->nPostIntsWorkItemScheduled++;
		MP_INC_REF(Adapter);				   
		NdisScheduleWorkItem(&Adapter->PostIntsWorkItem);	
	}

	SMSC_TRACE0(DBG_INT,"<-- NICPostInts\r\n");
}


/*++

Routine Description:

	This function stops further submissions on the interrupt pipe, 
	cancels all pending URBs and waits for their completion.
	
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	
Return value:

	VOID
--*/

NIC_STATUS
NICStopInterruptPipe(
	PMP_ADAPTER Adapter
	)
{ 
	BOOLEAN 		  bDone=TRUE;
	LONG			  nWaitCount = 0;
	NIC_STATUS		  Status=NIC_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_SHUT_DOWN,"--> NICStopInterruptPipe\r\n");
	
	if (Adapter->IsIntReady)
	 	NdisWaitEvent(&Adapter->IsIntsCompletionWorkItemReady,0);

	// Since it can be called from MP_Halt when MP_Initialize fails
	if(!MP_TEST_FLAG(Adapter, fMP_INT_RESOURCE_ALLOCATED)){
		Status=NIC_STATUS_FAILURE;
		goto Exit_NICStopInterruptPipe;
	}

	MP_CLEAR_FLAG(Adapter, fMP_POST_INTS);
	
	if (!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) {
		// Try to stop them before the completion in one shot.
		Adapter->Interrupt.PipeTransferAborted=TRUE;
		if ( ! Adapter->UsbFuncs->lpAbortPipeTransfers(Adapter->Interrupt.hPipe,USB_NO_WAIT))
		{
			SMSC_TRACE0(DBG_ERROR,"Interrupt Pipe Abort failed !!!\r\n");		
		}
	}
	
	// Wait for all int completions and ICBs to make it back to the free list.
	while(TRUE)
	{
		bDone = TRUE;
	   
		
		NdisAcquireSpinLock(&Adapter->IntLock);

		//
		// Are all the interrupt trancastions completed ?
		//		
		if(Adapter->nBusyInt + Adapter->nProcessedInt + Adapter->nCompletionInt)
		{
			SMSC_TRACE1(DBG_INT,"nBusyInt = %d\r\n", Adapter->nBusyInt);
			SMSC_TRACE1(DBG_INT,"nProcessedInt = %d\r\n", Adapter->nProcessedInt);
			SMSC_TRACE1(DBG_INT,"nCompletionInt = %d\r\n", Adapter->nCompletionInt);
			bDone = FALSE;
		}
		NdisReleaseSpinLock(&Adapter->IntLock); 

		if(bDone)
		{
			break;	 
		}

		if((++nWaitCount % NIC_1_SEC_TIMEOUT) == 0)
		{
			SMSC_TRACE1(DBG_ERROR,"Int Cancel timed out after %d second!!! COMMITING LEAK !!!\r\n", NIC_1_SEC_TIMEOUT/1000);
			SMSC_TRACE1(DBG_ERROR, "   Adapter->nBusyInt = %d\r\n", Adapter->nBusyInt);
			SMSC_TRACE1(DBG_ERROR, "   Adapter->nProcessedInt = %d\r\n", Adapter->nProcessedInt);
			SMSC_TRACE1(DBG_ERROR,"Adapter->nCompletionInt = %d\r\n", Adapter->nCompletionInt);
			Status=NIC_STATUS_FAILURE;
			ASSERT(FALSE);		 
			break;	
		}

		NdisMSleep(NDIS_SLEEP_1_MS);		
	}

	
Exit_NICStopInterruptPipe:
	
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- NICStopInterruptPipe\r\n");
	return Status;
}

/*++

Routine Description:

   Workitem to service interrupts. This workitme is scheduled in the 
   interrupt pipe completion routine each time we receive an interrupt 
   with status other than "receive data in receive fifo" 
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - Pointer to our Adapter

Return Value:

	None.

--*/
VOID
NICIntStatWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	DWORD			Value32;
	PICB			pICB=NULL;
		   
	SMSC_TRACE0(DBG_INT,"--> NICIntWorkItemCallBack\r\n");

	// Read the interrupt Status again now and process the sources.
	if (NICReadRegister(Adapter, ETH_INT_STS, &Value32) != NIC_STATUS_SUCCESS)
		goto Exit_NICIntStatWorkItemCallBack;
	
	if (Value32 & (~(ETH_INT_STS_GPIO_)))
	{
		SMSC_TRACE1(DBG_ERROR, "Interrupt Status = 0x%08lx\r\n", Value32); 

		if (Value32 & ETH_INT_STS_MACRTO_)
		{
			SMSC_TRACE0(DBG_ERROR, "   ETH_INT_STS_MACRTO_: MAC Reset Timeout, Should never happen !\r\n"); 
		}
		if (Value32 & ETH_INT_STS_TX_STOP_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TX_STOP_: Tx Stopped\r\n");
		}
		if (Value32 & ETH_INT_STS_RX_STOP_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_RX_STOP_: Rx Stopped\r\n");
		}
		if (Value32 & ETH_INT_STS_PHY_INT_)
		{
			SMSC_TRACE0(DBG_INT,"	ETH_INT_STS_PHY_INT_: PHY int\r\n");
		}
		if (Value32 & ETH_INT_STS_TXE_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TXE_	: TX Error, will also cause pipe stall. Handled by USB Error recovery.\r\n");
		}
		if (Value32 & ETH_INT_STS_TDFU_)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TDFU_: TX Underrun, Should never happen !\r\n");
		}
		if (Value32 & ETH_INT_STS_TDFO_	)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_TDFO_: TX Overrun, Should never happen !\r\n");
		}
		if (Value32 & ETH_INT_STS_RXDF_	)
		{
			SMSC_TRACE0(DBG_ERROR,"   ETH_INT_STS_RXDF_: Rx Dropped frame. \r\n");
		}
	}

	NICWriteRegister(Adapter, ETH_INT_STS, &Value32);
	
Exit_NICIntStatWorkItemCallBack:

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one.
	//
	InterlockedExchange(&Adapter->IsIntStatWorkItemQueued, FALSE);

	MP_DEC_REF(Adapter);	 
	
	SMSC_TRACE0(DBG_INT,"<-- NICIntWorkItemCallBack\r\n");		  
}

/*++

Routine Description:

	This routine sends a interrupt request to the USB
		
Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	pICB	-  Pointer to the ICB block 


Return Value:

	NT status code

--*/
NTSTATUS
NICPostIntRequest(
	PMP_ADAPTER Adapter,
	PICB	pICB
	)
{

//	PDEVICE_OBJECT	 TargetDeviceObject = Adapter->TargetDeviceObject;
	int count=0;
			  
	SMSC_TRACE1(DBG_INT,"--> NICPostIntRequest (pICB = %p)\r\n", pICB);

	if (!ACCEPT_IO(Adapter)) {
		pICB->Ref = 1;
		SMSC_TRACE0(DBG_ERROR,"NICPostIntRequest(),Device is not ready for I/O request!!!\r\n");
		return STATUS_UNSUCCESSFUL;
	}
		
	ResetEvent(pICB->SyncEvent);
	pICB->Ref = 1;
	pICB->dwErr= IssueInterruptTransfer( Adapter->UsbFuncs,
									   Adapter->Interrupt.hPipe,
									   NICIntRequestCompletion,
									   pICB,
									   (USB_IN_TRANSFER|USB_SHORT_TRANSFER_OK),
									   pICB->pData, 0,
									   INT_BUFFER_SIZE,
									  &pICB->dwTransferLength,
									   0,
									  &pICB->dwUsbErr);
			
	 SetEvent(pICB->SyncEvent);

//	 ASSERT(!(!pICB->dwErr && !pICB->dwUsbErr));
	 if ( (ERROR_SUCCESS ==  pICB->dwErr) && (USB_NO_ERROR ==  pICB->dwUsbErr)	) 
	 {

		SMSC_TRACE2(DBG_ERROR,"Interrupt Request, NULL Transfer,  pICB->dwErr:%d  pICB->dwUsbErr:0x%x)\r\n",  pICB->dwErr,  pICB->dwUsbErr);
		return STATUS_UNSUCCESSFUL;
			   
	 }

		
	 if ((ERROR_SUCCESS != pICB->dwErr)) {
		SMSC_TRACE2(DBG_ERROR,"IssueInterruptTransfer ERROR, dwErr= 0x%x, dwUsbErr=0x%x\r\n",pICB->dwErr,pICB->dwUsbErr);
		return STATUS_UNSUCCESSFUL;
	 }
	
	
	SMSC_TRACE0(DBG_INT,"<-- NICPostIntRequest\r\n");

	return STATUS_SUCCESS;
}


/*++

Routine Description:

	Completion routine for the Int request. This routine
	put the ICB into IntcompletionList and open another workitem to process it
		
Arguments:

	pICB -	pointer to the context structure (ICB) 

Return Value:

	STATUS_SUCCESS
	
--*/
DWORD
NICIntRequestCompletion(
	PICB   pICB
	)
{

	PMP_ADAPTER Adapter = pICB->Adapter;
	
	SMSC_TRACE0(DBG_INT,"--> NICIntRequestCompletion \r\n");

	NdisAcquireSpinLock(&Adapter->IntLock);

	// Remove it from the busy list 
	// and put it in the completion List
	RemoveEntryList(&pICB->List);	
	NdisInterlockedDecrement(&Adapter->nBusyInt);
	ASSERT(Adapter->nBusyInt >= 0);

	// Put it in the completion list.
	InsertTailList(&Adapter->IntCompletionList, &pICB->List);
	NdisInterlockedIncrement(&Adapter->nCompletionInt);
	ASSERT(Adapter->nCompletionInt <= Adapter->MaxICBs);

	SMSC_TRACE1(DBG_INT,"ICB %p removed from it's busy list and inserted into IntCompletionList\r\n", pICB);

	NdisReleaseSpinLock(&Adapter->IntLock);

	#ifdef USE_THREAD
		SetEvent(Adapter->IntThreadEvent);
	#else
		#ifdef USE_WORK_ITEM_NO_OUT
		 NdisSetEvent(&Adapter->IntsWorkItemReady);
		#else
		
		if (InterlockedExchange(&Adapter->IsIntsCompletionWorkItemQueued, TRUE) == FALSE)
		{
			Adapter->nIntsCompletionWorkItemScheduled++;
			MP_INC_REF(Adapter);
			NdisScheduleWorkItem(&Adapter->IntsCompletionWorkItem);
		}
		#endif
	#endif
	
	SMSC_TRACE0(DBG_INT,"<-- NICIntRequestCompletion\r\n");

	return STATUS_SUCCESS;
}

/*++

Routine Description:

	This routine process the ICBs in the IntCompletionList 
		
Arguments:

	Context  point to the adapter
Return Value:

	STATUS_SUCCESS
	
--*/

NTSTATUS
IntsCompletionWorkItemCallBack(
   PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	PMP_ADAPTER Adapter = Context;
	PICB pICB;
	BOOLEAN	bIntFailed = FALSE;
	ULONG	bytesRead;
	ULONG	dwInt;
	USB_ERROR	UsbErr;
	USB_TRANSFER hTransfer;
	BOOL dwErr=FALSE;
	
	SMSC_TRACE0(DBG_INT,"--> IntsCompletionWorkItemCallback. \r\n");
	
	NdisWaitEvent(&Adapter->IsIntsCompletionWorkItemReady, 0);
	NdisResetEvent(&Adapter->IsIntsCompletionWorkItemReady);

	InterlockedExchange(&Adapter->IsIntsCompletionWorkItemQueued, FALSE);


	while (!IsListEmpty(&Adapter->IntCompletionList)) {
	
		NdisAcquireSpinLock(&Adapter->IntLock);
		pICB = (PICB) RemoveHeadList(&Adapter->IntCompletionList);
		ASSERT(pICB && (pICB != (PICB) &Adapter->IntCompletionList));// cannot be NULL and list cannot be empty
		NdisInterlockedDecrement(&Adapter->nCompletionInt);
		ASSERT(Adapter->nCompletionInt >= 0);
		NdisReleaseSpinLock(&Adapter->IntLock);

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//wait for the IssueInterruptTransfer() to complete before we process the interrupt transfer	handler
		WaitForSingleObject(pICB->SyncEvent, 100);
		ResetEvent(pICB->SyncEvent);

	
		if (pICB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"Interrupt Request has NULL Transfer,  hTransfer=0x%x \r\n", pICB->dwUsbErr);	
		}


		hTransfer=(USB_TRANSFER)pICB->dwUsbErr;
		dwErr=GetTransferStatus(Adapter->UsbFuncs, hTransfer, &pICB->dwTransferLength,&UsbErr);


		if (hTransfer) {
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}

		bytesRead = pICB->dwTransferLength;
	
		if( !(USB_NO_ERROR == UsbErr) || !dwErr || Adapter->Interrupt.PipeTransferAborted)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
					UsbErr,
					dwErr,
					hTransfer);
						
			bIntFailed = TRUE;
		}
		else if (bytesRead != 4)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request invalid size = %d , Data Buffer: 0x%p, data= 0x%x\r\n", bytesRead, pICB->pData, *(pICB->pData));		
			bIntFailed = TRUE;
		}
	

		if (bIntFailed)
		{

			Adapter->nIntsCompletedWithError++; 	   

			//
			// Clear the flag to prevent any more ints from being
			// posted to the target device.
			//
			MP_CLEAR_FLAG(Adapter, fMP_POST_INTS);

			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) && \
				(!Adapter->Interrupt.PipeTransferAborted) &&   \
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			{

				SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
										UsbErr,
										dwErr,
										hTransfer);
			
				if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE) 
				{
						Adapter->nUsbErrorRecoveryWorkItemScheduled++;
						MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
				}
			}		
	   
		
		}
		else
		{
			// We got a good Interrupt transaction completion, count it.
			Adapter->nIntsCompletedSuccessfully++;
			dwInt = *(DWORD*) pICB->pData;
			SMSC_TRACE2(DBG_INT,"Int %d bytes = 0x%08lx\r\n", bytesRead, dwInt);
		}
 
		// Let's see if we should actually process it.
		if (MP_TEST_FLAG(Adapter, fMP_POST_INTS))	// Checking for this is enough since only NICStartInterruptPipe sets it
		{
			if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
			{
				if ((Adapter->StopBulkInOnRxEmpty) && (dwInt & INT_ENP_RFHF_)) 
				{
				MP_SET_FLAG(Adapter, fMP_POST_READS);
	
				// For performance, instead of scheduling reads at the end of
				// every int completion that showed data is available, we will 
				// do it only when the number of outstanding ICBs goes below 
				// Adapter->LowWatermarkICBs.
//				if((!Adapter->LowWatermarkICBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkICBs))
				{
					NICPostReads(Adapter);
				}
			}

			// We'll schedule an Int Status workitem if the interrupt itself
			// needs processing
			if (dwInt & ~(INT_ENP_RFHF_ | ETH_INT_STS_GPIOS_))
			{
				// If a workitem already scheduled don't queue another.
				if (InterlockedExchange(&Adapter->IsIntStatWorkItemQueued, TRUE) == FALSE)
				{
					Adapter->nIntStatWorkItemScheduled++;
					MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->IntStatWorkItem);
				}
			}
		}
	}

	// We're done with this ICB
	NdisAcquireSpinLock(&Adapter->IntLock);

	// Put it in the Processed list.
	InsertTailList(&Adapter->IntProcessedList, &pICB->List);
	NdisInterlockedIncrement(&Adapter->nProcessedInt);
	ASSERT(Adapter->nProcessedInt <= Adapter->MaxICBs);

	SMSC_TRACE1(DBG_INT,"ICB %p inserted into IntProcessedList\r\n", pICB);

	NdisReleaseSpinLock(&Adapter->IntLock); 

	if (NdisInterlockedDecrement(&pICB->Ref) == 0) 
	{
		NICFreeICB(pICB);
	}
	else
	{
		// This should never happen
		SMSC_TRACE0(DBG_ERROR, "NICIntRequestCompletion finds ICB with invalid Ref count!\r\n");			 
	}
}
	
	MP_DEC_REF(Adapter); 
	NdisSetEvent(&Adapter->IsIntsCompletionWorkItemReady);
	SMSC_TRACE0(DBG_INT,"<-- IntCompletionWorkItemCallback\r\n");

	return STATUS_SUCCESS;
}



/*++

Routine Description:

	This routine process the ICBs in the IntCompletionList 
		
Arguments:

	Context  point to the adapter
Return Value:

	STATUS_SUCCESS
	
--*/

NTSTATUS
IntsCompletionCallBack(
   PMP_ADAPTER Adapter
	)
{

	PICB pICB;
	BOOLEAN	bIntFailed = FALSE;
	ULONG	bytesRead;
	ULONG	dwInt;
	USB_ERROR	UsbErr;
	USB_TRANSFER hTransfer;
	BOOL dwErr=FALSE;
	
	SMSC_TRACE0(DBG_INT,"--> IntsCompletionWorkItemCallback. \r\n");
	
	if (Adapter->MPInitializeFail ||(!MP_TEST_FLAG(Adapter, fMP_INT_RESOURCE_ALLOCATED)))
				return STATUS_SUCCESS;

	while (!IsListEmpty(&Adapter->IntCompletionList)) {
	
		NdisAcquireSpinLock(&Adapter->IntLock);
		pICB = (PICB) RemoveHeadList(&Adapter->IntCompletionList);
		ASSERT(pICB && (pICB != (PICB) &Adapter->IntCompletionList));// cannot be NULL and list cannot be empty
		NdisInterlockedDecrement(&Adapter->nCompletionInt);
		ASSERT(Adapter->nCompletionInt >= 0);
		NdisReleaseSpinLock(&Adapter->IntLock);

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//wait for the IssueInterruptTransfer() to complete before we process the interrupt transfer	handler
		WaitForSingleObject(pICB->SyncEvent, 100);
		ResetEvent(pICB->SyncEvent);

	
		if (pICB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"Interrupt Request has NULL Transfer,  hTransfer=0x%x \r\n", pICB->dwUsbErr);	
		}


		hTransfer=(USB_TRANSFER)pICB->dwUsbErr;
		dwErr=GetTransferStatus(Adapter->UsbFuncs, hTransfer, &pICB->dwTransferLength,&UsbErr);


		if (hTransfer) {
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}

		bytesRead = pICB->dwTransferLength;
	
		if( !(USB_NO_ERROR == UsbErr) || !dwErr || Adapter->Interrupt.PipeTransferAborted)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
					UsbErr,
					dwErr,
					hTransfer);
						
			bIntFailed = TRUE;
		}
		else if (bytesRead != 4)
		{
			SMSC_TRACE3(DBG_ERROR, "Int request invalid size = %d , Data Buffer: 0x%p, data= 0x%x\r\n", bytesRead, pICB->pData, *(pICB->pData));		
			bIntFailed = TRUE;
		}
	

		if (bIntFailed)
		{

			Adapter->nIntsCompletedWithError++; 	   

			//
			// Clear the flag to prevent any more ints from being
			// posted to the target device.
			//
			MP_CLEAR_FLAG(Adapter, fMP_POST_INTS);

			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) && \
				(!Adapter->Interrupt.PipeTransferAborted) &&   \
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			{

				SMSC_TRACE3(DBG_ERROR, "Int request failed UsbErr = %x, dwErr = %x, hTransfer=IF0x%x\r\n", 
										UsbErr,
										dwErr,
										hTransfer);
			
				if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE) 
				{
						Adapter->nUsbErrorRecoveryWorkItemScheduled++;
						MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
				}
			}		
	   
		
		}
		else
		{
			// We got a good Interrupt transaction completion, count it.
			Adapter->nIntsCompletedSuccessfully++;
			dwInt = *(DWORD*) pICB->pData;
			SMSC_TRACE2(DBG_INT,"Int %d bytes = 0x%08lx\r\n", bytesRead, dwInt);
		}
 
		// Let's see if we should actually process it.
		if (MP_TEST_FLAG(Adapter, fMP_POST_INTS))	// Checking for this is enough since only NICStartInterruptPipe sets it
		{
			if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
			{
				if ((Adapter->StopBulkInOnRxEmpty) && (dwInt & INT_ENP_RFHF_)) 
				{
				MP_SET_FLAG(Adapter, fMP_POST_READS);
	
				// For performance, instead of scheduling reads at the end of
				// every int completion that showed data is available, we will 
				// do it only when the number of outstanding ICBs goes below 
				// Adapter->LowWatermarkICBs.
//				if((!Adapter->LowWatermarkICBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkICBs))
				{
					NICPostReads(Adapter);
				}
			}

			// We'll schedule an Int Status workitem if the interrupt itself
			// needs processing
			if (dwInt & ~(INT_ENP_RFHF_ | ETH_INT_STS_GPIOS_))
			{
				// If a workitem already scheduled don't queue another.
				if (InterlockedExchange(&Adapter->IsIntStatWorkItemQueued, TRUE) == FALSE)
				{
					Adapter->nIntStatWorkItemScheduled++;
					MP_INC_REF(Adapter);				   
					NdisScheduleWorkItem(&Adapter->IntStatWorkItem);
				}
			}
		}
	}

	// We're done with this ICB
	NdisAcquireSpinLock(&Adapter->IntLock);

	// Put it in the Processed list.
	InsertTailList(&Adapter->IntProcessedList, &pICB->List);
	NdisInterlockedIncrement(&Adapter->nProcessedInt);
	ASSERT(Adapter->nProcessedInt <= Adapter->MaxICBs);

	SMSC_TRACE1(DBG_INT,"ICB %p inserted into IntProcessedList\r\n", pICB);

	NdisReleaseSpinLock(&Adapter->IntLock); 

	if (NdisInterlockedDecrement(&pICB->Ref) == 0) 
	{
		NICFreeICB(pICB);
	}
	else
	{
		// This should never happen
		SMSC_TRACE0(DBG_ERROR, "NICIntRequestCompletion finds ICB with invalid Ref count!\r\n");			 
	}
}
	
	SMSC_TRACE0(DBG_INT,"<-- IntCompletionWorkItemCallback\r\n");

	return STATUS_SUCCESS;
}


/*++

Routine Description:

	This routine reinitializes the ICB block and puts it back
	into the IntFreeList for reuse. Assumes the ICB was in the 
	process list
		
Arguments:

	pICB	  - pointer to ICB block   

Return Value:

	VOID

--*/
VOID 
NICFreeICB(
	IN PICB pICB
	)
{
	PMP_ADAPTER Adapter = pICB->Adapter;

	SMSC_TRACE1(DBG_INT,"--> NICFreeICB %p\r\n", pICB);
	
	ASSERT(!pICB->Ref); // should be 0
	ASSERT(pICB->Adapter); // shouldn't be NULL

	NdisAcquireSpinLock(&Adapter->IntLock);
	
	RemoveEntryList(&pICB->List);
	NdisInterlockedDecrement(&Adapter->nProcessedInt);
	ASSERT(Adapter->nProcessedInt >= 0);
   
	SMSC_TRACE1(DBG_INT,"ICB %p removed from the Processed List and inserted into IntFreeList\r\n", pICB);
	InsertTailList(&Adapter->IntFreeList, &pICB->List);

	NdisReleaseSpinLock(&Adapter->IntLock); 

	if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED))
	{
		// For performance, unless the caller specifically requested 
		// a workitem, instead of scheduling reads at the end of
		// every read completion, we will do it only when the number of 
		// outstanding ICBs goes below Adapter->LowWatermarkICBs.
		if(
			((!Adapter->LowWatermarkICBs) || (Adapter->nBusyInt <= Adapter->LowWatermarkICBs))
			&& MP_TEST_FLAG(Adapter, fMP_POST_INTS)
		  )
		{
			NICPostInts(Adapter);
		}
	}

	SMSC_TRACE1(DBG_INT,"<-- NICFreeICB %d\r\n", Adapter->nBusyInt);
}

#endif

/*++

Routine Description:
	
	This is the Linkcheck Timer Callback function. 

Arguments:

	FunctionContext - Pointer to our Adapter
	Others			- Not used 

Return Value:

	None.

--*/

VOID 
NICLinkCheck(
	IN PVOID  SystemSpecific1,
	IN PVOID  FunctionContext,
	IN PVOID  SystemSpecific2,
	IN PVOID  SystemSpecific3
	)
{
	PMP_ADAPTER Adapter = (PMP_ADAPTER) FunctionContext;

	UNREFERENCED_PARAMETER(SystemSpecific1);
	UNREFERENCED_PARAMETER(SystemSpecific2);
	UNREFERENCED_PARAMETER(SystemSpecific3);
	NdisAcquireSpinLock(&Adapter->LCSpinlock);

	SMSC_TRACE0(DBG_LINK,"--> NICLinkCheck\r\n");

	if (MP_IS_READY(Adapter))
	{
		
		if (InterlockedExchange(&Adapter->IsLinkCheckWorkItemQueued, TRUE) == FALSE)
		{
			MP_SET_FLAG(Adapter, fMP_LINK_CHECK_IN_PROGRESS);
			Adapter->nLinkCheckWorkItemScheduled++;
			SMSC_TRACE0(DBG_LINK,"	 Callback Scheduled\r\n");
			MP_INC_REF(Adapter);				   
			NdisScheduleWorkItem(&Adapter->LinkCheckWorkItem);	 
		}
		
		else
		{
			SMSC_TRACE0(DBG_LINK,"	 NOT Scheduling callback, one still pending.\r\n");
		}
		
	}

	// Signal any thread waiting for the Timer Cancellation that we're done.
	NdisSetEvent(&Adapter->LinkCheckCalledEvent);

	NdisReleaseSpinLock(&Adapter->LCSpinlock);

	SMSC_TRACE0(DBG_LINK,"<-- NICLinkCheck\r\n");
}

__inline
VOID
NICUpdateStatCounters(
    IN PMP_ADAPTER Adapter,
    IN PUCHAR pStats
    )
/*++

Routine Description:

    This routine updates Tx/Rx statistic counters.

Arguments:

    Adapter - pointer to our Adapter
    pStats - Pointer to stat. counters array

Return Value:

--*/
{
    PLAN7500_STATS   pMacStats = (PLAN7500_STATS) pStats;

    // --fixme so many counters are not reported in this routine. Needs update
    
    if (Adapter->ReinitStatCounter == FALSE)
    {
        // Good Tx (20Bit Counter)
        if (pMacStats->TxGoodUnicastFrames >= Adapter->MacStats.TxGoodUnicastFrames)
            Adapter->TxGoodUnicastFrames += (ULONG64)(pMacStats->TxGoodUnicastFrames - Adapter->MacStats.TxGoodUnicastFrames);
        else
            Adapter->TxGoodUnicastFrames += (ULONG64)(pMacStats->TxGoodUnicastFrames + (0xFFFFFUL - Adapter->MacStats.TxGoodUnicastFrames) + 1UL);

        if (pMacStats->TxGoodBroadcastFrames >= Adapter->MacStats.TxGoodBroadcastFrames)
            Adapter->TxGoodBroadcastFrames += (ULONG64)(pMacStats->TxGoodBroadcastFrames - Adapter->MacStats.TxGoodBroadcastFrames);
        else
            Adapter->TxGoodBroadcastFrames += (ULONG64)(pMacStats->TxGoodBroadcastFrames + (0xFFFFFUL - Adapter->MacStats.TxGoodBroadcastFrames) + 1UL);

        if (pMacStats->TxGoodMulticastFrames >= Adapter->MacStats.TxGoodMulticastFrames)
            Adapter->TxGoodMulticastFrames += (ULONG64)(pMacStats->TxGoodMulticastFrames - Adapter->MacStats.TxGoodMulticastFrames);
        else
            Adapter->TxGoodMulticastFrames += (ULONG64)(pMacStats->TxGoodMulticastFrames + (0xFFFFFUL - Adapter->MacStats.TxGoodMulticastFrames) + 1UL);

        Adapter->TransmitsGood = (Adapter->TxGoodUnicastFrames + Adapter->TxGoodBroadcastFrames + Adapter->TxGoodMulticastFrames);
        
        // Tx Excess Collisions (20Bit Counter)
        if (pMacStats->TxExcessiveCollisionErrors >= Adapter->MacStats.TxExcessiveCollisionErrors)
            Adapter->TxAbortExcessCollisions += (pMacStats->TxExcessiveCollisionErrors - Adapter->MacStats.TxExcessiveCollisionErrors);
        else
            Adapter->TxAbortExcessCollisions += (pMacStats->TxExcessiveCollisionErrors + (0xFFFFFUL - Adapter->MacStats.TxExcessiveCollisionErrors) + 1UL);

        // Tx Late Collisions (20Bit Counter)
        if (pMacStats->TxLateCollisionErrors >= Adapter->MacStats.TxLateCollisionErrors)
            Adapter->TxLateCollisions += (pMacStats->TxLateCollisionErrors - Adapter->MacStats.TxLateCollisionErrors);
        else
            Adapter->TxLateCollisions += (pMacStats->TxLateCollisionErrors + (0xFFFFFUL - Adapter->MacStats.TxLateCollisionErrors) + 1UL);

        // Tx Carrier Errors (20Bit Counter)
        if (pMacStats->TxCarrierErrors >= Adapter->MacStats.TxCarrierErrors)
            Adapter->TxLostCRS += (pMacStats->TxCarrierErrors - Adapter->MacStats.TxCarrierErrors);
        else
            Adapter->TxLostCRS += (pMacStats->TxCarrierErrors + (0xFFFFFUL - Adapter->MacStats.TxCarrierErrors) + 1UL);

        // Tx Single Collision (20Bit Counter)
        if (pMacStats->TxSingleCollisions >= Adapter->MacStats.TxSingleCollisions)
            Adapter->OneRetry += (pMacStats->TxSingleCollisions - Adapter->MacStats.TxSingleCollisions);
        else
            Adapter->OneRetry += (pMacStats->TxSingleCollisions + (0xFFFFFUL - Adapter->MacStats.TxSingleCollisions) + 1UL);

        // Tx Multiple Collision (20Bit Counter)
        if (pMacStats->TxMultipleCollisions >= Adapter->MacStats.TxMultipleCollisions)
            Adapter->MoreThanOneRetry += (pMacStats->TxMultipleCollisions - Adapter->MacStats.TxMultipleCollisions);
        else
            Adapter->MoreThanOneRetry += (pMacStats->TxMultipleCollisions + (0xFFFFFUL - Adapter->MacStats.TxMultipleCollisions) + 1UL);

        // Tx Excessive Deferral (20Bit Counter)
        if (pMacStats->TxExcessiveDeferralErrors >= Adapter->MacStats.TxExcessiveDeferralErrors)
            Adapter->TransmitFailuresOther += (pMacStats->TxExcessiveDeferralErrors - Adapter->MacStats.TxExcessiveDeferralErrors);
        else
            Adapter->TransmitFailuresOther += (pMacStats->TxExcessiveDeferralErrors + (0xFFFFFUL - Adapter->MacStats.TxExcessiveDeferralErrors) + 1UL);

        // Tx Pause (20Bit Counter)
        if (pMacStats->TxPauseFrames >= Adapter->MacStats.TxPauseFrames)
            Adapter->TxPauseFrames += (pMacStats->TxPauseFrames - Adapter->MacStats.TxPauseFrames);
        else
            Adapter->TxPauseFrames += (pMacStats->TxPauseFrames + (0xFFFFFUL - Adapter->MacStats.TxPauseFrames) + 1UL);

        // Good Rx (20Bit Counter)
        if (pMacStats->RxGoodUnicastFrames >= Adapter->MacStats.RxGoodUnicastFrames)
            Adapter->RxGoodUnicastFrames += (ULONG64)(pMacStats->RxGoodUnicastFrames - Adapter->MacStats.RxGoodUnicastFrames);
        else
            Adapter->RxGoodUnicastFrames += (ULONG64)(pMacStats->RxGoodUnicastFrames + (0xFFFFFUL - Adapter->MacStats.RxGoodUnicastFrames) + 1UL);

        if (pMacStats->RxGoodBroadcastFrames >= Adapter->MacStats.RxGoodBroadcastFrames)
            Adapter->RxGoodBroadcastFrames += (ULONG64)(pMacStats->RxGoodBroadcastFrames - Adapter->MacStats.RxGoodBroadcastFrames);
        else
            Adapter->RxGoodBroadcastFrames += (ULONG64)(pMacStats->RxGoodBroadcastFrames + (0xFFFFFUL - Adapter->MacStats.RxGoodBroadcastFrames) + 1UL);

        if (pMacStats->RxGoodMulticastFrames >= Adapter->MacStats.RxGoodMulticastFrames)
            Adapter->RxGoodMulticastFrames += (ULONG64)(pMacStats->RxGoodMulticastFrames - Adapter->MacStats.RxGoodMulticastFrames);
        else
            Adapter->RxGoodMulticastFrames += (ULONG64)(pMacStats->RxGoodMulticastFrames + (0xFFFFFUL - Adapter->MacStats.RxGoodMulticastFrames) + 1UL);

        Adapter->ReceivesGood = (Adapter->RxGoodUnicastFrames + Adapter->RxGoodBroadcastFrames + Adapter->RxGoodMulticastFrames);

        // Rx Crc (20Bit Counter)
        if (pMacStats->RxCrcErrors >= Adapter->MacStats.RxCrcErrors)  
            Adapter->RcvCrcErrors += (pMacStats->RxCrcErrors - Adapter->MacStats.RxCrcErrors);
        else
            Adapter->RcvCrcErrors += (pMacStats->RxCrcErrors + (0xFFFFFUL - Adapter->MacStats.RxCrcErrors) + 1UL);

        // Rx Alignment (20Bit Counter)
        if (pMacStats->RxAlignmentErrors >= Adapter->MacStats.RxAlignmentErrors)
            Adapter->RcvAlignmentErrors += (pMacStats->RxAlignmentErrors - Adapter->MacStats.RxAlignmentErrors);
        else
            Adapter->RcvAlignmentErrors += (pMacStats->RxAlignmentErrors + (0xFFFFFUL - Adapter->MacStats.RxAlignmentErrors) + 1UL);

        // Rx Resources (20Bit Counter)
        if (pMacStats->RxDroppedFrames >= Adapter->MacStats.RxDroppedFrames)
            Adapter->RcvResourceErrors += (pMacStats->RxDroppedFrames - Adapter->MacStats.RxDroppedFrames);
        else
            Adapter->RcvResourceErrors += (pMacStats->RxDroppedFrames + (0xFFFFFUL - Adapter->MacStats.RxDroppedFrames) + 1UL);

        // Rx Runts (20Bit Counter)
        if (pMacStats->RxRuntFrameErrors >= Adapter->MacStats.RxRuntFrameErrors)
            Adapter->RcvRuntErrors += (pMacStats->RxRuntFrameErrors - Adapter->MacStats.RxRuntFrameErrors);
        else
            Adapter->RcvRuntErrors += (pMacStats->RxRuntFrameErrors + (0xFFFFFUL - Adapter->MacStats.RxRuntFrameErrors) + 1UL);

        // Rx TooLong (20Bit Counter)
        if (pMacStats->RxJabberErrors >= Adapter->MacStats.RxJabberErrors)
            Adapter->RcvFrameTooLongErrors += (pMacStats->RxJabberErrors - Adapter->MacStats.RxJabberErrors);
        else
            Adapter->RcvFrameTooLongErrors += (pMacStats->RxJabberErrors + (0xFFFFFUL - Adapter->MacStats.RxJabberErrors) + 1UL);

        // Rx Late Collisions (20Bit Counter)
        // -fixme it may not be a right candidate for Late Collision Error Counter
        if (pMacStats->RxFragmentErrors >= Adapter->MacStats.RxFragmentErrors)
            Adapter->RcvLateCollisionErrors += (pMacStats->RxFragmentErrors - Adapter->MacStats.RxFragmentErrors);
        else
            Adapter->RcvLateCollisionErrors += (pMacStats->RxFragmentErrors + (0xFFFFFUL - Adapter->MacStats.RxFragmentErrors) + 1UL);
    }
    
    Adapter->ReinitStatCounter = FALSE;
    
    Adapter->MacStats.TxGoodUnicastFrames = pMacStats->TxGoodUnicastFrames;
    Adapter->MacStats.TxGoodBroadcastFrames = pMacStats->TxGoodBroadcastFrames;
    Adapter->MacStats.TxGoodMulticastFrames = pMacStats->TxGoodMulticastFrames;
    Adapter->MacStats.TxExcessiveCollisionErrors = pMacStats->TxExcessiveCollisionErrors;
    Adapter->MacStats.TxLateCollisionErrors = pMacStats->TxLateCollisionErrors;
    Adapter->MacStats.TxCarrierErrors = pMacStats->TxCarrierErrors;
    Adapter->MacStats.TxSingleCollisions = pMacStats->TxSingleCollisions;
    Adapter->MacStats.TxMultipleCollisions = pMacStats->TxMultipleCollisions;
    Adapter->MacStats.TxExcessiveDeferralErrors = pMacStats->TxExcessiveDeferralErrors;
    Adapter->MacStats.TxPauseFrames = pMacStats->TxPauseFrames;

    Adapter->MacStats.RxGoodUnicastFrames = pMacStats->RxGoodUnicastFrames;
    Adapter->MacStats.RxGoodBroadcastFrames = pMacStats->RxGoodBroadcastFrames;
    Adapter->MacStats.RxGoodMulticastFrames = pMacStats->RxGoodMulticastFrames;
    Adapter->MacStats.RxCrcErrors = pMacStats->RxCrcErrors;
    Adapter->MacStats.RxAlignmentErrors = pMacStats->RxAlignmentErrors;
    Adapter->MacStats.RxDroppedFrames = pMacStats->RxDroppedFrames;
    Adapter->MacStats.RxRuntFrameErrors = pMacStats->RxRuntFrameErrors;
    Adapter->MacStats.RxJabberErrors = pMacStats->RxJabberErrors;
    Adapter->MacStats.RxFragmentErrors = pMacStats->RxFragmentErrors;
}


/*++

Routine Description:

   Workitem to check link state. This workitem is scheduled in the link check 
   timer callback function 
	  
Arguments:

	WorkItem - Pointer to workitem
	
	Context - Pointer to our Adapter

Return Value:

	None.

--*/
VOID
NICLinkCheckWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	NIC_STATUS		Status;
	DWORD			Value32;
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;

	SMSC_TRACE0(DBG_LINK,"--> NICLinkCheckWorkItemCallBack\r\n");

	if (!MP_TEST_FLAG(Adapter, fMP_PERIODIC_LINK_CHECK) || !MP_IS_READY(Adapter))
		goto Exit_NICLinkCheckWorkItemCallBack;

	if(!NICIsCardPresent(Adapter))
		goto Exit_NICLinkCheckWorkItemCallBack;

/*
	Status = NICReadStats(Adapter, Stats);

	
	if (Status == NIC_STATUS_SUCCESS)
	{
		NICUpdateStatCounters(Adapter, Stats);
	}
	else
	{
		SMSC_TRACE1(DBG_LINK,"NICLinkCheckWorkItemCallBack. Error reading stats (0x%08lx)\r\n", Status);
	}
*/
		
	// Check Link

	if(Adapter->PerformingNWAY)
	{
		SMSC_TRACE1(DBG_LINK,"NWAY elapsed time = %d seconds since first check\r\n", Adapter->PerformingNWAYTimerCounter);
		if(NICIsNWAYComplete(Adapter) || (Adapter->PerformingNWAYTimerCounter == 10))
		{
			Status = NICServicePhyInterrupt(Adapter); // returns Adapter->PerformingNWAY = FALSE only if linked
			Adapter->PerformingNWAY = FALSE;

			NdisAcquireSpinLock(&Adapter->RequestLock);
/*
			if (Adapter->HardwareStatus == NdisHardwareStatusReset)
			{
				SMSC_TRACE1(DBG_RESET,"Reset status1= 0x%x \r\n", (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY);
				NdisMResetComplete(
									Adapter->AdapterHandle, 
									(!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY,
									FALSE
								  );
			}
*/
			Adapter->HardwareStatus = NdisHardwareStatusReady;
			NdisReleaseSpinLock(&Adapter->RequestLock); 
		}

		Adapter->PerformingNWAYTimerCounter++;
	}
     else
    {
        // Is there still a pending Oid Request we have not completed and maybe could ?
        if ((!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && (Adapter->SecondsElapsedAfterMPInitializeExit > (Adapter->AnegCompleteTiomeout-1)))
        {
            // Complete it !
            NICCompleteOidQueryIfPending(Adapter);
        }
    }


	// If we are disconnected and EDPD is enabled, see if we need to 
	// come out of EDPD
	if ((MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Adapter->PhyLinkPowerDown == TRUE))
	{
		Status = NICReadPhy(Adapter, PHY_MOD_CTL_STS, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICLinkCheckWorkItemCallBack;
		SMSC_TRACE1(DBG_LINK,"PHY_MOD_CTL_STS=0x%08lx\r\n", Value32);

		if (Value32 & PHY_MOD_CTL_STS_ENERGYON_)
		{
			SMSC_TRACE0(DBG_LINK,"Energy Detected !!! Turning OFF EDPD\r\n");

			Value32 &= ~PHY_MOD_CTL_STS_EDPWRDOWN_;
			Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Value32);
			if (Status != NIC_STATUS_SUCCESS)
				goto Exit_NICLinkCheckWorkItemCallBack;

			Adapter->PhyLinkPowerDown = FALSE;

			if (!Adapter->PhyIsExternal)
			{
				// Restore Auto MDIX
				Status = NICWritePhy(Adapter, PHY_MOD_CTL_STS, Adapter->SaveAutoMdix);
				if(Status != NIC_STATUS_SUCCESS)
					goto Exit_NICLinkCheckWorkItemCallBack;

				Status = NICWritePhy(Adapter, PHY_EXT_CTL_STS, Adapter->SaveMdixState);
                		if(Status != NIC_STATUS_SUCCESS)
                    			goto Exit_NICLinkCheckWorkItemCallBack;
						
				SMSC_TRACE0(DBG_LINK,"AMDIX Restored\r\n");
			}

			// If we were doing NWAY turn it ON now
			if(Adapter->MediaType < MEDIA_TYPE_10HD_FORCED)
			{
				Status = NICWritePhy(  
										Adapter, 
										PHY_CTRL, 
                                        MII_CR_AUTO_NEG_EN_ |MII_CR_RESTART_AUTO_NEG_
									 );
				if (Status != NIC_STATUS_SUCCESS)
					goto Exit_NICLinkCheckWorkItemCallBack;
	
				SMSC_TRACE0(DBG_LINK,"NWAY Reenabled\r\n");
			}
		}
	}

	// The Phy is still powered down, skip all link checks 
	if (Adapter->PhyLinkPowerDown == TRUE)
		goto Exit_NICLinkCheckWorkItemCallBack;

	// Check to see if the media was reconnected to workaround problem of 
	//	no int generated on link reconnection when NWAY is off.

	if(
		!Adapter->PerformingNWAY &&
		(Adapter->MediaType < MEDIA_TYPE_10HD_FORCED) 
	  )
	{
		Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICLinkCheckWorkItemCallBack;

		Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
		if (Status != NIC_STATUS_SUCCESS)
			goto Exit_NICLinkCheckWorkItemCallBack;

		if( (MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Value32 & MII_SR_LINK_STATUS_) )
		{
			MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
			NICIndicateMediaStatus(Adapter);
		}
	}

	// If the hardware status was not ready, and we are now then switch it to ready.

	NdisAcquireSpinLock(&Adapter->RequestLock);
	if(
		(!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) &&
		(
			(Adapter->HardwareStatus == NdisHardwareStatusInitializing) ||
			(Adapter->HardwareStatus == NdisHardwareStatusReset)
		)
	  )
	{
/*
		if (Adapter->HardwareStatus == NdisHardwareStatusReset)
		{
			SMSC_TRACE1(DBG_RESET,"Reset status2= 0x%x \r\n", (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY);
			NdisMResetComplete(
								Adapter->AdapterHandle, 
								(!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) ? NDIS_STATUS_SUCCESS : NDIS_STATUS_ADAPTER_NOT_READY,
								FALSE
							  );

		}
*/
		SMSC_TRACE0(DBG_LINK,"Changing Hardware Status (OID_GEN_HARDWARE_STATUS) to ready\r\n");
		Adapter->HardwareStatus = NdisHardwareStatusReady;
	}
	NdisReleaseSpinLock(&Adapter->RequestLock); 

	// If two seconds elapsed since returning from MPInitialize and we still
	// don't have link, indicate a media disconnect event.

	if (MP_TEST_FLAG(Adapter, fMP_DISCONNECTED) && 
		((Adapter->SecondsElapsedAfterMPInitializeExit == (Adapter->AnegCompleteTiomeout+1)) ||
		(Adapter->SecondsElapsedAfterMPResetExit == (Adapter->MpRstCompleteTimeout+1)))
	   )
	{

		NICIndicateMediaStatus(Adapter);
	}

	// If six seconds elapsed since returning from MP_Reset with PENDING status and we still
	// don't have link, indicate a reset completion with the adapter not ready.
/*
	NdisAcquireSpinLock(&Adapter->RequestLock);
	if (
		(Adapter->HardwareStatus == NdisHardwareStatusReset) &&
		(Adapter->SecondsElapsedAfterMPResetExit >= (Adapter->MpRstCompleteTimeout-1)) &&
		(MP_TEST_FLAG(Adapter, fMP_DISCONNECTED))
	   )
	{
		Adapter->HardwareStatus = NdisHardwareStatusNotReady;

		SMSC_TRACE1(DBG_RESET,"Reset status3= 0x%x \r\n",  NDIS_STATUS_ADAPTER_NOT_READY);
		NdisMResetComplete(
							Adapter->AdapterHandle, 
							NDIS_STATUS_ADAPTER_NOT_READY,
							FALSE
						  );


	}
	NdisReleaseSpinLock(&Adapter->RequestLock); 
*/
	// In case the phy does not support it or we miss the link 
	// disconnection interrupt do a simple check for link change
	Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICLinkCheckWorkItemCallBack;

	Status = NICReadPhy(Adapter, PHY_STATUS, &Value32);
	if (Status != NIC_STATUS_SUCCESS)
		goto Exit_NICLinkCheckWorkItemCallBack;

	if ((!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && !(Value32 & MII_SR_LINK_STATUS_) )
	{
		NICServicePhyInterrupt(Adapter);
	}
	else if ((MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Value32 & MII_SR_LINK_STATUS_) )
	{
		NICServicePhyInterrupt(Adapter);
	}

	// Is there still a pending Oid Request we have not completed and maybe could ?
	if (!Adapter->PerformingNWAY && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
	{
		// Complete it !
		if(MP_TEST_FLAG(Adapter, fMP_REQUEST_PENDING))
			NICCompleteOidQueryIfPending(Adapter);
	}

Exit_NICLinkCheckWorkItemCallBack:
	// Update counters

	Adapter->SecondsElapsedAfterMPInitializeExit++;
	// Take care of the case when the counter wraps around to 0.
	if (Adapter->SecondsElapsedAfterMPInitializeExit == 0)
	{
		Adapter->SecondsElapsedAfterMPInitializeExit = Adapter->AnegCompleteTiomeout;
	}

	Adapter->SecondsElapsedAfterMPResetExit++;
	// Take care of the case when the counter wraps around to 0.
	if (Adapter->SecondsElapsedAfterMPResetExit == 0)
	{
		Adapter->SecondsElapsedAfterMPResetExit = Adapter->MpRstCompleteTimeout;
	}

	MP_CLEAR_FLAG(Adapter, fMP_LINK_CHECK_IN_PROGRESS);

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one.
	//

	InterlockedExchange(&Adapter->IsLinkCheckWorkItemQueued, FALSE);

	MP_DEC_REF(Adapter);	 

	SMSC_TRACE0(DBG_LINK,"<-- NICLinkCheckWorkItemCallBack\r\n");
}

NIC_STATUS 
NICServicePhyInterrupt(
    PMP_ADAPTER Adapter
    )
/*++

Routine Description:

   This routine services the Phy Interrupt
      
Arguments:

    Adapter    - pointer to the MP_ADAPTER structure
    
Return Value:

    None.

--*/
{
    NIC_STATUS Status;
    DWORD Value32,Value32_2,CommonDenom, Temp32;

    // If by any chance we get here when using a serial PHY just go out
    // without attempting to talk to an MII PHY.

    // Reading BMS serves the purpose of seeing if we need to update 
    // link and speed in the device, and do media connection indication.
    // Actually read it twice, as sometimes the Link does not get updated
    // until the second read in some PHYs that latch the link down condition.

	SMSC_TRACE0(DBG_PHY,"--> NICServicePhyInterrupt()\r\n");

    Status = NICReadPhy(Adapter, PHY_STATUS, &Temp32);
    if (Status != NIC_STATUS_SUCCESS)
        goto Exit_NICServicePhyInterrupt;

    if( (Adapter->MediaType < MEDIA_TYPE_10HD_FORCED) && 
        (Temp32 & MII_SR_LINK_STATUS_) )
    {
        // If NWAY was in process then signal we are done
        Adapter->PerformingNWAY = FALSE;

        /* Get link speed */
        // First check for 1000Mbps
        Status = NICReadPhy(Adapter, PHY_1000T_CTRL, &Value32_2);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICServicePhyInterrupt;

        Status = NICReadPhy(Adapter, PHY_1000T_STATUS, &Value32);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICServicePhyInterrupt;

        if ((Value32_2 & CR_1000T_FD_CAPS_)  && // LAN98500 can only do 1000T-FD
            (Value32 & SR_1000T_LP_FD_CAPS_))
        {
            // Master/Slave state of the PHY
            if (Value32 & SR_1000T_MS_CONFIG_RES_) // Master
            {
                Adapter->LinkSpeed = LINK_SPEED_1000MBS_MASTER;
                SMSC_TRACE0(DBG_LINK,"Link Speed = 1000Mbs Master\r\n");
            }
            else //Slave
            {
                Adapter->LinkSpeed = LINK_SPEED_1000MBS_SLAVE;
                SMSC_TRACE0(DBG_LINK,"Link Speed = 1000Mbs Slave\r\n");
            }
            
            Adapter->LinkDuplex = LINK_DUPLEX_FULL;
            Adapter->PhyRxClkWorkaround = TRUE;
            SMSC_TRACE0(DBG_LINK,"Duplex = Full\r\n");
        }
        else
        {
            Status = NICReadPhy(Adapter, PHY_AUTONEG_ADV, &Value32_2);
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICServicePhyInterrupt;

            Status = NICReadPhy(Adapter, PHY_LP_ABILITY, &Value32);
            if (Status != NIC_STATUS_SUCCESS)
                goto Exit_NICServicePhyInterrupt;

            // Get the least common denominator
            CommonDenom = Value32 & Value32_2;

            // Set the flags before calling NICSetMediaType.
            Adapter->LinkSpeed = LINK_SPEED_10MBS;
            if(CommonDenom & (NWAY_AR_100TX_FD_CAPS_ | NWAY_AR_100TX_HD_CAPS_)) 
            {
                Adapter->LinkSpeed = LINK_SPEED_100MBS;
                SMSC_TRACE0(DBG_LINK,"Link Speed = 100Mbs\r\n");
            } 
            else 
            {
                if(CommonDenom & (NWAY_AR_10T_FD_CAPS_ | NWAY_AR_10T_HD_CAPS_))
                {
                    SMSC_TRACE0(DBG_LINK,"Link Speed = 10Mbs\r\n");
                }
                else
                {
                    SMSC_TRACE0(DBG_LINK,"Link Speed = Unknown\r\n");
                }
                if (Adapter->PhyRxClkWorkaround) {
                    MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
                    Status = NICInitializeMedia(Adapter);
                    Adapter->PhyRxClkWorkaround = FALSE;
                    return Status;
                }
            }

            /* Get link duplex mode */
            Adapter->LinkDuplex = LINK_DUPLEX_HALF;
            if(CommonDenom & (NWAY_AR_100TX_FD_CAPS_ | NWAY_AR_10T_FD_CAPS_)) 
            {
                Adapter->LinkDuplex = LINK_DUPLEX_FULL;
                SMSC_TRACE0(DBG_LINK,"Duplex = Full\r\n");
            } 
            else 
            {
                if(CommonDenom & (NWAY_AR_100TX_HD_CAPS_ | NWAY_AR_10T_HD_CAPS_))
                {
                    SMSC_TRACE0(DBG_LINK,"Duplex = Half\r\n");
                }
                else 
                {
                    SMSC_TRACE0(DBG_LINK,"Duplex = Unknown\r\n");
                }
            }
        }

        if ((Adapter->LinkDuplex == LINK_DUPLEX_FULL) && (Adapter->FlowControl != DISABLED))
        {
            if ((Adapter->LinkSpeed == LINK_SPEED_1000MBS_MASTER) ||
                (Adapter->LinkSpeed == LINK_SPEED_1000MBS_SLAVE))
            {
                Status = NICReadPhy(Adapter, PHY_AUTONEG_ADV, &Value32_2);
                if (Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICServicePhyInterrupt;

                Status = NICReadPhy(Adapter, PHY_LP_ABILITY, &Value32);
                if (Status != NIC_STATUS_SUCCESS)
                    goto Exit_NICServicePhyInterrupt;
            }
            
            // We are autonegotiating and if it is operating in full duplex, if 
            // Flow Control enabled determine it's options for pause generation 
            // and response according to IEEE 802.3 spec table 28B-3 in Annex 28B. // --fixme check the reference
            if (Value32_2 & NWAY_AR_PAUSE_)
            {
                // Local PAUSE set
                if (Value32_2 & NWAY_AR_ASM_DIR_)
                {
                    // Local PAUSE set, Local ASM_DIR set
                    if (Value32 & NWAY_LPAR_PAUSE_)
                    {
                        // Local PAUSE set, Local ASM_DIR set, Remote Pause Set, Remote ASM_DIR don't care
                        Adapter->bRespondToPause = TRUE;
                        Adapter->bGeneratePause = TRUE;
                    }
                    else
                    {
                        // Local PAUSE set, Local ASM_DIR set, Remote Pause clr
                        if (Value32 & NWAY_LPAR_ASM_DIR_)
                        {
                            // Local PAUSE set, Local ASM_DIR set, Remote Pause clr, Remote ASM_DIR set
                            Adapter->bRespondToPause = TRUE;
                            Adapter->bGeneratePause = FALSE;
                        }
                        else
                        {
                            // Local PAUSE set, Local ASM_DIR set, Remote Pause clr, Remote ASM_DIR clr
                            Adapter->bRespondToPause = FALSE;
                            Adapter->bGeneratePause = FALSE;
                        }
                    }
                }
                else
                {
                    // Local PAUSE set, Local ASM_DIR clr
                    if (Value32 & NWAY_LPAR_PAUSE_)
                    {
                        // Local PAUSE set, Local ASM_DIR clr, Remote Pause Set, Remote ASM_DIR don't care
                        Adapter->bRespondToPause = TRUE;
                        Adapter->bGeneratePause = TRUE;
                    }
                    else
                    {
                        // Local PAUSE set, Local ASM_DIR clr, Remote Pause clr, Remote ASM_DIR don't care
                        Adapter->bRespondToPause = FALSE;
                        Adapter->bGeneratePause = FALSE;
                    }
                }
            }
            else
            {
                // Local PAUSE Clr
                if (Value32_2 & NWAY_AR_ASM_DIR_)
                {
                    // Local PAUSE Clr, Local ASM_DIR set
                    if (Value32 & NWAY_LPAR_PAUSE_)
                    {
                        // Local PAUSE Clr, Local ASM_DIR set, Remote Pause Set
                        if (Value32 & NWAY_LPAR_ASM_DIR_)
                        {
                            // Local PAUSE Clr, Local ASM_DIR set, Remote Pause Set, Remote ASM_DIR Set
                            Adapter->bRespondToPause = FALSE;
                            Adapter->bGeneratePause = TRUE;
                        }
                        else
                        {
                            // Local PAUSE Clr, Local ASM_DIR set, Remote Pause Set, Remote ASM_DIR Clr
                            Adapter->bRespondToPause = FALSE;
                            Adapter->bGeneratePause = FALSE;
                        }
                    }
                    else
                    {
                        // Local PAUSE Clr, Local ASM_DIR set, Remote Pause Clr, Remote ASM_DIR don't care.
                        Adapter->bRespondToPause = FALSE;
                        Adapter->bGeneratePause = FALSE;
                    }
                }
                else
                {
                    // Local PAUSE Clr, Local ASM_DIR clr, Remote Don't care.
                    Adapter->bRespondToPause = FALSE;
                    Adapter->bGeneratePause = FALSE;
                }
            }
        }
        else
        {
            // Half duplex, just setup backoff and backpressure options 
            // according to init driver settings
            switch(Adapter->FlowControl)
            {
                case DISABLED:
                    Adapter->bGeneratePause = FALSE;
                    Adapter->bRespondToPause = FALSE;
					break;

                case FC_SYMM:
                    Adapter->bRespondToPause = TRUE;
                    Adapter->bGeneratePause = TRUE;
                    break;

                case FC_ASYM_GEN:
                    Adapter->bGeneratePause = TRUE;
                    Adapter->bRespondToPause = FALSE;
                    break;

                case FC_ASYM_RESP:
                    Adapter->bGeneratePause = FALSE;
                    Adapter->bRespondToPause = TRUE;
                    break;
            }
        }

        Status = NICSetMediaType(Adapter);
        if (Status != NIC_STATUS_SUCCESS)
            goto Exit_NICServicePhyInterrupt;
    }

    // Signal Media connection status change if necessary, but only if NWAY is not
    // currently in progress.
    if (!Adapter->PerformingNWAY)
    {
        if( (MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (Temp32 & MII_SR_LINK_STATUS_) )
        {
            MP_CLEAR_FLAG(Adapter, fMP_DISCONNECTED);
            NICIndicateMediaStatus(Adapter);
        }

        if( (!MP_TEST_FLAG(Adapter, fMP_DISCONNECTED)) && (!(Temp32 & MII_SR_LINK_STATUS_)) )
        {
            MP_SET_FLAG(Adapter, fMP_DISCONNECTED);
            NICIndicateMediaStatus(Adapter);
        }

        
    }

    Status = NIC_STATUS_SUCCESS;

Exit_NICServicePhyInterrupt:
	SMSC_TRACE0(DBG_PHY,"<-- NICServicePhyInterrupt()\r\n");
    
    return Status;
}

