/*++

	Copyright (c) 2009-2010 SMSC. All rights reserved.

	The information contained in this file is confidential and proprietary to
	Standard Microsystems Corporation.	No part of this file may be reproduced
	or distributed, in any form or by any means for any purpose, without the
	express written permission of Standard Microsystems Corporation.


Module Name:

	receive.c

Abstract:

		This module contains miniport functions for handling receive
		packets and other helper routines called by these miniport functions.

Revision History:
	01/19/2010	Initial version.

Notes:

--*/


#include "smsc7500.h"


/*++

Routine Description:

	NDIS Miniport entry point called whenever protocols are done with
	a packet that we had indicated up and they had queued up for returning
	later.

Arguments:

	MiniportAdapterContext	  - pointer to MP_ADAPTER structure
	Packet	  - packet being returned.

Return Value:

	None.

--*/
VOID
MPReturnPacket(
	IN NDIS_HANDLE	MiniportAdapterContext,
	IN PNDIS_PACKET Packet)
{
	PRCB pRCB = NULL;
	PMP_ADAPTER Adapter;

	pRCB = *(PRCB *)Packet->MiniportReservedEx;

	SMSC_TRACE2(DBG_RX,"--> MPReturnPacket (RCB = %p, Packet = %p)\r\n", pRCB, Packet);

	Adapter = pRCB->Adapter;

	ASSERT(Adapter);

	Adapter->nRxPacketsReturned++;

	if(NdisInterlockedDecrement(&pRCB->Ref) == 0)
	{
		NICFreeRCB(pRCB);
	}

	SMSC_TRACE0(DBG_RX,"<-- MPReturnPacket\r\n");
}

/*++

Routine Description:

   Workitem to post all the free read requests to the target
   driver. This workitem is scheduled from the MiniportInitialize
   worker routine during start (Bulk only mode), from the NICFreeRCB
   whenever the outstanding read requests to the target driver goes
   below the watermark. And also from the completion routine of the
   interrupt transaction if that says there's data available (usb
   friendly mode)

Arguments:

	WorkItem - Pointer to workitem

	Context - Pointer to our Adapter

Return Value:

	None.

--*/

VOID
 NICPostReadsWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem,
	PVOID Context
	)
{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	NTSTATUS		ntStatus;
	PRCB			pRCB=NULL;

	SMSC_TRACE0(DBG_RX,"--> NICPostReadsWorkItemCallBack\r\n");

	//
	// If another thread has this same workitem running already
	// wait for it finish.
	//
	NdisWaitEvent(&Adapter->ReadWorkItemReady, 0);
	NdisResetEvent(&Adapter->ReadWorkItemReady);

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one. Clear it before processing to avoid
	// race condition which can leave the device in a state where
	// no further bulk in transactions are submitted, specially
	// in Bulk Only mode since the int pipe cannot kick start it.
	//
	InterlockedExchange(&Adapter->IsReadWorkItemQueued, FALSE);

	//
	// Now lets actually post some bulk in reads.
	//
	NdisAcquireSpinLock(&Adapter->BulkInLock);

	while(MP_IS_READY_AND_LINKED_UP(Adapter) && !IsListEmpty(&Adapter->BulkInFreeList) && MP_TEST_FLAG(Adapter, fMP_POST_READS))
	{
		pRCB = (PRCB) RemoveHeadList(&Adapter->BulkInFreeList);
		// no need to initialize the list pointers since we are immediately storing in another queue

		ASSERT(pRCB && (pRCB != (PRCB) &Adapter->BulkInFreeList));// cannot be NULL and list cannot be empty

		//
		// Insert the RCB in the recv busy queue
		//
		InsertTailList(&Adapter->BulkInBusyList, &pRCB->List);
		NdisInterlockedIncrement(&Adapter->nBusyRecv);
		ASSERT(Adapter->nBusyRecv <= Adapter->MaxRCBs);

		NdisReleaseSpinLock(&Adapter->BulkInLock);

		Adapter->nBulkInPosted++;
		ntStatus = NICPostReadRequest(Adapter, pRCB);

		if (!NT_SUCCESS ( ntStatus ) )
		{
			SMSC_TRACE1(DBG_ERROR, "NICPostReadRequest failed %x\r\n", ntStatus);

			// NICFreeRCB expects it in the Processed queue, so put it there
			// before making the call.

			NdisAcquireSpinLock(&Adapter->BulkInLock);

			// Remove it from the busy list 
			RemoveEntryList(&pRCB->List);
			NdisInterlockedDecrement(&Adapter->nBusyRecv);
			ASSERT(Adapter->nBusyRecv >= 0);

			// Put it in the Processed list.
			InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedRecv);
			ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

			SMSC_TRACE1(DBG_RX,"RCB %p removed from the busy and inserted into BulkInProcessedList\r\n", pRCB);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
			{
				Adapter->SubmitRCBFail=TRUE;
				NICFreeRCB(pRCB);
			}
			else
			{
				// This should never happen
				SMSC_TRACE0(DBG_ERROR,"NICPostReadsWorkItemCallBack finds RCB with invalid Ref count!\r\n");
				SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
			}

			NdisAcquireSpinLock(&Adapter->BulkInLock);
			break;
		}

		NdisAcquireSpinLock(&Adapter->BulkInLock);
	}

	NdisReleaseSpinLock(&Adapter->BulkInLock);

	//
	// If another thread has this same workitem ready to run again
	// but waiting for us to finish, signal it to go ahead.
	//
	NdisSetEvent(&Adapter->ReadWorkItemReady);

	MP_DEC_REF(Adapter);

	SMSC_TRACE0(DBG_RX,"<-- NICPostReadsWorkItemCallBack\r\n");
}


/*++

Routine Description:

   Workitem to post all the free read requests to the target
   driver. This workitem is scheduled from the MiniportInitialize
   worker routine during start (Bulk only mode), from the NICFreeRCB
   whenever the outstanding read requests to the target driver goes
   below the watermark. And also from the completion routine of the
   interrupt transaction if that says there's data available (usb
   friendly mode)

Arguments:

	WorkItem - Pointer to workitem

	Context - Pointer to our Adapter

Return Value:

	None.

--*/

VOID
 NICPostReadsCallBack(
	PMP_ADAPTER 	Adapter
	)
{
	NTSTATUS		ntStatus;
	PRCB			pRCB=NULL;

	SMSC_TRACE0(DBG_RX,"--> NICPostReadsCallBack\r\n");

	
	//
	// Now lets actually post some bulk in reads.
	//
	NdisAcquireSpinLock(&Adapter->BulkInLock);

	while(MP_IS_READY_AND_LINKED_UP(Adapter) && !IsListEmpty(&Adapter->BulkInFreeList) && MP_TEST_FLAG(Adapter, fMP_POST_READS))
	{
		pRCB = (PRCB) RemoveHeadList(&Adapter->BulkInFreeList);
		// no need to initialize the list pointers since we are immediately storing in another queue

		ASSERT(pRCB && (pRCB != (PRCB) &Adapter->BulkInFreeList));// cannot be NULL and list cannot be empty

		//
		// Insert the RCB in the recv busy queue
		//
		InsertTailList(&Adapter->BulkInBusyList, &pRCB->List);
		NdisInterlockedIncrement(&Adapter->nBusyRecv);
		ASSERT(Adapter->nBusyRecv <= Adapter->MaxRCBs);

		NdisReleaseSpinLock(&Adapter->BulkInLock);

		Adapter->nBulkInPosted++;
		ntStatus = NICPostReadRequest(Adapter, pRCB);

		if (!NT_SUCCESS ( ntStatus ) )
		{
			SMSC_TRACE1(DBG_ERROR, "NICPostReadRequest failed %x\r\n", ntStatus);

			// NICFreeRCB expects it in the Processed queue, so put it there
			// before making the call.

			NdisAcquireSpinLock(&Adapter->BulkInLock);

			// Remove it from the busy list 
			RemoveEntryList(&pRCB->List);
			NdisInterlockedDecrement(&Adapter->nBusyRecv);
			ASSERT(Adapter->nBusyRecv >= 0);

			// Put it in the Processed list.
			InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedRecv);
			ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

			SMSC_TRACE1(DBG_RX,"RCB %p removed from the busy and inserted into BulkInProcessedList\r\n", pRCB);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
			{
				Adapter->SubmitRCBFail=TRUE;
				NICFreeRCB(pRCB);
			}
			else
			{
				// This should never happen
				SMSC_TRACE0(DBG_ERROR,"NICPostReadsWorkItemCallBack finds RCB with invalid Ref count!\r\n");
				SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
			}

			NdisAcquireSpinLock(&Adapter->BulkInLock);
			break;
		}

		NdisAcquireSpinLock(&Adapter->BulkInLock);
	}

	NdisReleaseSpinLock(&Adapter->BulkInLock);

	SMSC_TRACE0(DBG_RX,"<-- NICPostReadsCallBack\r\n");
}







/*++

Routine Description:

	Starts the Receive path.

	Note that if we are operating in interrupt mode we defer
	starting the bulk in requests for now (will start when Interrupt pipe signals
	receive data available).

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return Value:

	NIC_STATUS_FAILURE if reset pipe failed, NIC_STATUS_SUCCESS otherwise

--*/
NIC_STATUS
NICStartRxPath(
	PMP_ADAPTER Adapter
	)
{
	DWORD Value32,Temp32;
	NIC_STATUS nStatus = NIC_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_RX,"--> NICStartRxPath\r\n");

	// We need to reset the pipe to make sure the data toggles are in sync. 
	if(!NICResetPipe(Adapter, Adapter->BulkIn))
	{
		SMSC_TRACE0(DBG_ERROR,"Bulk In Pipe Reset failed !!! Aborting Bulk In Start sequence.\r\n");
		nStatus = NIC_STATUS_FAILURE;
		goto Exit_NICStartRxPath;
	}
	
	Adapter->BulkIn.PipeTransferAborted=0;
	Adapter->SubmitRCBFail=FALSE;
	
	Adapter->BulkInExpectedSeq = 0;
	Adapter->BulkInSubmitedSeq = 0;
	MP_CLEAR_FLAG(Adapter, fMP_IGNORE_READS);

	if (!Adapter->StopBulkInOnRxEmpty)
	{
		SMSC_TRACE0(DBG_RX,"USB BULK only mode\r\n");
		NICStartBulkIn(Adapter);
		
	}
	else
	{
		SMSC_TRACE0(DBG_RX,"USB bandwidth friendly mode, Skipping starting the Bulk in pipe\r\n");
		// We don't start Bulk Ins now, but do leave the workitem in a startable state.
		NdisSetEvent(&Adapter->ReadWorkItemReady);
		NdisSetEvent(&Adapter->BulkInResumeWorkItemReady);
		NdisSetEvent(&Adapter->BulkInCompletionWorkItemReady);
	}

	// --fixme This should be default value in ASIC and no need to reprogram
    	nStatus = NICWriteRegister(Adapter, ETH_FCT_RX_FIFO_END, &(Adapter->RxFifoEnd));
    	if (nStatus != NIC_STATUS_SUCCESS)
        	goto Exit_NICStartRxPath;

    	// Enable Rx FIFO, also clear Rx FIFO Disabled flag
    	nStatus = NICReadRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
		Temp32= Value32 | ETH_FCT_RX_CTL_EN_ | ETH_FCT_RX_CTL_DIS_;
    	nStatus |= NICWriteRegister(Adapter, ETH_FCT_RX_CTL, &Temp32 );
    	if (nStatus != NIC_STATUS_SUCCESS)
        	goto Exit_NICStartRxPath;

    	// Enable Rx at MAC, also clear Rx Disabled flag
   		nStatus = NICReadRegister(Adapter, ETH_MAC_RX, &Value32);
    
    	// Clear then set Maximum Frame Size
    	Value32 &= (~ETH_MAC_RX_MAX_SIZE_);
    	Value32 |= (((Adapter->MaxFrameSize + 4UL) & 0x3FFFUL) << ETH_MAC_RX_MAX_SIZE_SHIFT_); // + 4 byte FCS
    	Value32 |= ETH_MAC_RX_RXEN_ | ETH_MAC_RX_RXD_;
    	nStatus |= NICWriteRegister(Adapter, ETH_MAC_RX, &Value32);
    	if (nStatus != NIC_STATUS_SUCCESS)
        	goto Exit_NICStartRxPath;

		Adapter->IsRecvReady=TRUE;

Exit_NICStartRxPath:
	SMSC_TRACE0(DBG_RX,"<-- NICStartRxPath\r\n");

	return (nStatus);
}

/*++

Routine Description:

	Posts all the free read requests on USB.

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return Value:

	None.

--*/
void
NICStartBulkIn(
	PMP_ADAPTER Adapter
	)
{
	SMSC_TRACE0(DBG_RX,"--> NICStartBulkIn\r\n");

	MP_SET_FLAG(Adapter, fMP_POST_READS);

	NdisSetEvent(&Adapter->ReadWorkItemReady);
	NdisSetEvent(&Adapter->BulkInResumeWorkItemReady);
	NdisSetEvent(&Adapter->BulkInCompletionWorkItemReady);
	
	SMSC_TRACE0(DBG_RX,"NICStartBulkIn, schedule ReadWorkItem\r\n");
	Adapter->nReadWorkItemScheduled++;
	MP_INC_REF(Adapter);
	NdisScheduleWorkItem(&Adapter->ReadWorkItem);

//	NICPostReads(Adapter);

	SMSC_TRACE0(DBG_RX,"<-- NICStartBulkIn\r\n");
}

/*++

Routine Description:

   Post all the free read requests on USB if the flag to do so is enabled.

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return Value:

	None.

--*/
void
NICPostReads(
	PMP_ADAPTER Adapter
	)
{
	SMSC_TRACE0(DBG_RX,"--> NICPostReads\r\n");
	
	NICPostReadsCallBack(Adapter);
   
	SMSC_TRACE0(DBG_RX,"<-- NICPostReads\r\n");
}


/*++

Routine Description:

	This routine sends a read request  to the USB to get
	the incoming network packet from the device.

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure
	pRCB	-  Pointer to the RCB block.


Return Value:

	NT status code

--*/

NTSTATUS
NICPostReadRequest(
	PMP_ADAPTER Adapter,
	PRCB	pRCB
	)
{
  
	SMSC_TRACE1(DBG_RX,"--> NICPostReadRequest (RCB = %p)\r\n", pRCB);

	if (!ACCEPT_IO(Adapter)) {
		pRCB->Ref = 1;
		SMSC_TRACE0(DBG_ERROR,"NICPostReadRequest(),Device is not ready for I/O request!!!\r\n");
		return STATUS_UNSUCCESSFUL;
	}

	ResetEvent(pRCB->SyncEvent);
	pRCB->Ref = 1;
	pRCB->dwErr= IssueBulkTransfer( Adapter->UsbFuncs,
									Adapter->BulkIn.hPipe,
									NICReadRequestCompletion,
									pRCB,
									(USB_IN_TRANSFER|USB_SHORT_TRANSFER_OK),
									pRCB->pData, 0,
									Adapter->RxBurstMax,
									&pRCB->dwTransferLength,
									0,
									&pRCB->dwUsbErr);


	SetEvent(pRCB->SyncEvent);

//	ASSERT(!(!pRCB->dwErr && !pRCB->dwUsbErr));

	if ( (ERROR_SUCCESS ==	pRCB->dwErr) && (USB_NO_ERROR ==  pRCB->dwUsbErr)  )
	{
		SMSC_TRACE2(DBG_ERROR,"Read Request, NULL Transfer,  pRCB->dwErr:%d  pRCB->dwUsbErr:0x%x)\r\n",  pRCB->dwErr,	pRCB->dwUsbErr);
		return STATUS_UNSUCCESSFUL;
	}


	if ((ERROR_SUCCESS != pRCB->dwErr)) {
		SMSC_TRACE2(DBG_ERROR,"Issue Bulk In Transfer ERROR, dwErr= 0x%x, dwUsbErr=0x%x\r\n",pRCB->dwErr,pRCB->dwUsbErr);
		return STATUS_UNSUCCESSFUL;
	}
	
	pRCB->seq = Adapter->BulkInSubmitedSeq;

	Adapter->BulkInSubmitedSeq = (Adapter->BulkInSubmitedSeq + 1) % Adapter->MaxRCBs;

	SMSC_TRACE0(DBG_RX,"<-- NICPostReadRequest\r\n");

	return STATUS_SUCCESS;
}


/*++

Routine Description:

	Processes all the packets in the BulkInWait list

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return Value:

	VOID
--*/

VOID
NICProcessBulkInWaitList(
	PMP_ADAPTER Adapter
	)
{
	DWORD	FrameLen;
	PRCB	pRCB;
	BOOLEAN bStop;

	NdisAcquireSpinLock(&Adapter->RCBProcessLock);

	SMSC_TRACE0(DBG_RX,"--> NICProcessBulkInWaitList\r\n");

	while (TRUE)
	{
		bStop = FALSE;


		// Grab the one at the head of the waiting list
		NdisAcquireSpinLock(&Adapter->BulkInLock);

		pRCB = (PRCB) RemoveHeadList(&Adapter->BulkInWaitList);
		if (pRCB == (PRCB) &Adapter->BulkInWaitList)
		{
			// Wait list empty, get out
			NdisReleaseSpinLock(&Adapter->BulkInLock);
			break;
		}

		// Make sure if we use RemoveEntryList on this RCB later we don't corrupt the queue.
		NdisInitializeListHead(&pRCB->List);

		NdisInterlockedDecrement(&Adapter->nWaitingRecv);
		ASSERT(Adapter->nWaitingRecv >= 0);

		SMSC_TRACE1(DBG_RX,"RCB %p removed from the wait list\r\n", pRCB);

		NdisReleaseSpinLock(&Adapter->BulkInLock);

		 FrameLen = ( *((DWORD *)pRCB->pData) & RX_CMD_A_LEN_ );

		if (
			(Adapter->RxOnePktperRcb) || (FrameLen == (pRCB->ulSize - 8UL))
		   )
		{
			// Single Receive Packet
			NICProcessSingleReceivePacket(pRCB);
		}
		else
		{
			// Multiple Receive Packets
//			if (Adapter->RxBurstCap)
//			{
				// Integer number of packets in RCB, no partial packets
				bStop = !NICProcessMultipleReceivePackets(pRCB, pRCB->ulSize - pRCB->ulCopied);
//			}
//			else
//			{
				// Non integer number of packets in RCB, could have partial packets
//				bStop = !NICProcessByteStreamReceivePackets(pRCB, pRCB->ulSize - pRCB->ulCopied);
//			}
		}

		if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
		{
			ASSERT(&pRCB->Ref >= 0);
			NICFreeRCB(pRCB);
		}
		

		ASSERT(&pRCB->Ref >= 0);

		if(bStop)
		{
			break;
		}
	}

	SMSC_TRACE0(DBG_RX,"<-- NICProcessBulkInWaitList\r\n");

	NdisReleaseSpinLock(&Adapter->RCBProcessLock);
}

/*++

Routine Description:

	Workitem to resume processing pending Bulk In Completions when
	split packets had been exhausted in multipacket mode. This workitem 
	is scheduled from the NICFreeRCB routine whenever there were pending
	USB RCBs (bulk in completions) pending and a split packet is released
	
Arguments:

		WorkItem - Pointer to workitem

	Context - Pointer to our Adapter

Return Value:

		None.

--*/

VOID
NICBulkInResumeWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem,
	PVOID Context
	)
{
	PMP_ADAPTER 	Adapter = (PMP_ADAPTER)Context;
	
	SMSC_TRACE0(DBG_RX,"--> NICBulkInResumeWorkItemCallBack\r\n");

	NdisWaitEvent(&Adapter->BulkInResumeWorkItemReady, 0);
	NdisResetEvent(&Adapter->BulkInResumeWorkItemReady);

	//
	// Clear the flag to let the WorkItem structure be reused for
	// scheduling another one. Clear it before processing to avoid
	// race condition.
	//
	InterlockedExchange(&Adapter->IsBulkInResumeWorkItemQueued, FALSE);

	// Now process all the RCBs we've collected in the wait list
	NICProcessBulkInWaitList(Adapter);

	NdisSetEvent(&Adapter->BulkInResumeWorkItemReady);
	MP_DEC_REF(Adapter);

	SMSC_TRACE0(DBG_RX,"<-- NICBulkInResumeWorkItemCallBack\r\n");
}

/*++

Routine Description:

		Completion routine for the read request. This routine	put RCBs in 
		the completion list and schedule a workItem to proccess.

Arguments:

	pRCB -	 pointer to the context structure (RCB) passed when the 
				RCB was submitted 

Return Value:

	STATUS_SUCCESS

--*/

NTSTATUS
NICReadRequestCompletion(
	PRCB			pRCB
	)
{
	PMP_ADAPTER Adapter = pRCB->Adapter;
	

	SMSC_TRACE1(DBG_RX,"--> NICReadRequestCompletion (RCB = %p)\r\n", pRCB);

	NdisAcquireSpinLock(&Adapter->BulkInLock);
	
	RemoveEntryList(&pRCB->List);
	NdisInterlockedDecrement(&Adapter->nBusyRecv);
	ASSERT(Adapter->nBusyRecv >= 0);
	
	InsertTailList(&Adapter->BulkInCompletionList, &pRCB->List);
	NdisInterlockedIncrement(&Adapter->nCompletionRecv);
	ASSERT(Adapter->nCompletionRecv <= Adapter->MaxRCBs);

	SMSC_TRACE1(DBG_RX,"RCB %p removed from it's current list (busy if in any) and inserted into BulkInCompletionList\r\n", pRCB);

	NdisReleaseSpinLock(&Adapter->BulkInLock);


	#ifdef USE_THREAD

	SetEvent(Adapter->RecvThreadEvent);

	#else
		#ifdef USE_WORK_ITEM_NO_OUT
		 NdisSetEvent(&Adapter->RecvWorkItemReady);
		#else
		
		if (InterlockedExchange(&Adapter->IsBulkInCompletionWorkItemQueued, TRUE) == FALSE)
		{
			Adapter->nBulkIncompletionWorkItemScheduled++;
			MP_INC_REF(Adapter);
			NdisScheduleWorkItem(&Adapter->BulkInCompletionWorkItem);
		}
		#endif
	#endif

	SMSC_TRACE0(DBG_RX,"<-- NICReadRequestCompletion\r\n");
	return STATUS_SUCCESS;
}



/*++

Routine Description:

	WorkItem used to proccess BulkInCompletionList
	
Arguments:

		WorkItem - Pointer to workitem
		Context -	pointer to the context structure (RCB)
	
Return Value:

	STATUS_SUCCESS
--*/


NTSTATUS
BulkInCompletionWorkItemCallBack(
	PNDIS_WORK_ITEM WorkItem, 
	PVOID Context
	)
{
	
	PMP_ADAPTER Adapter = Context;
	PRCB		   pRCB;
	BOOLEAN bIndicateReceive, bBulkInFailed = FALSE;
	ULONG	seq; 
	USB_ERROR	UsbErr;
	USB_TRANSFER hTransfer;
	BOOL dwErr=FALSE;
	int timeout=0;

	SMSC_TRACE0(DBG_RX,"--> BulkInCompletionWorkItemCallBack \r\n");

		
	NdisWaitEvent(&Adapter->BulkInCompletionWorkItemReady, 0);
	NdisResetEvent(&Adapter->BulkInCompletionWorkItemReady);

	InterlockedExchange(&Adapter->IsBulkInCompletionWorkItemQueued, FALSE);

	 NdisAcquireSpinLock(&Adapter->BulkInLock);
	while (!IsListEmpty(&Adapter->BulkInCompletionList) ) {

	   
		pRCB = (PRCB) RemoveHeadList(&Adapter->BulkInCompletionList);
		ASSERT(pRCB && (pRCB != (PRCB) &Adapter->BulkInCompletionList));// cannot be NULL and list cannot be empty
		NdisInterlockedDecrement(&Adapter->nCompletionRecv);
		ASSERT(Adapter->nCompletionRecv >= 0);
		NdisReleaseSpinLock(&Adapter->BulkInLock);
		
		seq= pRCB->seq;

		if (pRCB->ulCopied != 0)
			SMSC_TRACE1(DBG_ERROR,"BulkIn completion, pRCB->ulCopied = %d != 0,  \r\n",pRCB->ulCopied);
		ASSERT(pRCB->ulCopied == 0);

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//wait for Bulk In request to finish, so we can proccess a valid Bulk In transfer handler
		WaitForSingleObject(pRCB->SyncEvent, 100);	
		ResetEvent(pRCB->SyncEvent);

		if (pRCB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"ReadRequest has NULL Transfer,  hTransfer=0x%x \r\n", pRCB->dwUsbErr);
		}

		hTransfer=(USB_TRANSFER)pRCB->dwUsbErr;

		dwErr=GetTransferStatus(Adapter->UsbFuncs, hTransfer, &pRCB->dwTransferLength,&UsbErr);

		if (hTransfer){
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}

		if( !(USB_NO_ERROR == UsbErr) || !dwErr || (Adapter->BulkIn.PipeTransferAborted) || !hTransfer)
		{
			bBulkInFailed = TRUE;
			Adapter->nBulkInCompletedWithError++;
			SMSC_TRACE5(DBG_ERROR,"ReadRequest failed dwErr = %x, dwUsbErr = %x, Aborted= %d, hTransfer=RF0x%x,RCB = %p \r\n",	\
										   dwErr, UsbErr, Adapter->BulkIn.PipeTransferAborted, hTransfer,pRCB);
		
			//
			// Clear the flags to prevent any more reads from being
			// posted to the target device.
			//
			MP_CLEAR_FLAG(Adapter, fMP_POST_READS);
			MP_SET_FLAG(Adapter, fMP_IGNORE_READS);

			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) && \
				(!Adapter->BulkIn.PipeTransferAborted) &&\
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			{
			 
				SMSC_TRACE2(DBG_ERROR,"NICReadRequestCompletion. USB Error !!! dwErr = %d, URB_Status = 0x%08lx\r\n",dwErr, UsbErr);

				// If device still present and workitem not already scheduled queue recovery.
				if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE)
				{
						Adapter->nUsbErrorRecoveryWorkItemScheduled++;
						MP_INC_REF(Adapter);
						NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
				}
			}

		}


		if (bBulkInFailed)
		{


			NdisAcquireSpinLock(&Adapter->BulkInLock);

			// Put it in the Processed list.
			InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedRecv);
			ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

			SMSC_TRACE1(DBG_RX,"RCB %p removed from it's current list (busy if in any) and inserted into BulkInProcessedList\r\n", pRCB);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
			{
				NICFreeRCB(pRCB);
			}
			else
			{
				// This should never happen
				SMSC_TRACE0(DBG_ERROR,"NICReadRequestCompletion(place1) finds RCB with invalid Ref count!\r\n");
				SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
			}
		 NdisAcquireSpinLock(&Adapter->BulkInLock);
			continue;
		}

		pRCB->ulSize = pRCB->dwTransferLength;
		
		NdisAcquireSpinLock(&Adapter->BulkInLock);

		// HOLD THIS LOCK UNTIL THE RCB IS INSERTED INTO THE NEXT APPPLICABLE
		// QUEUE. THIS IS SPECIALLY VITAL TO PREVENT GETTING OUT OF ORDER IN
		// THE WAIT LIST !!!

		ASSERT(Adapter->BulkInCompletionArray[seq] == NULL);
		Adapter->BulkInCompletionArray[seq] = pRCB;
		SMSC_TRACE2(DBG_RX,"RCB %p stored in array position %d\r\n", pRCB, seq);

		if (seq != Adapter->BulkInExpectedSeq)
		{
			SMSC_TRACE3(DBG_ERROR,"RCB %p out of order ! seq = %d, Expected seq = %d\r\n", pRCB, seq, Adapter->BulkInExpectedSeq);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
			{
				// For performance, instead of scheduling reads at the end of
				// every read completion, we will do it only when the number of
				// outstanding RCBs goes below Adapter->LowWatermarkRCBs.
//				if(((!Adapter->LowWatermarkRCBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkRCBs))
//					&& MP_TEST_FLAG(Adapter, fMP_POST_READS))
				if(MP_TEST_FLAG(Adapter, fMP_POST_READS))
				{
					NICPostReads(Adapter);
				}
			}
		 NdisAcquireSpinLock(&Adapter->BulkInLock);
			continue;
		}

		do
		{
			bIndicateReceive = FALSE;

			pRCB = Adapter->BulkInCompletionArray[Adapter->BulkInExpectedSeq];

			ASSERT(pRCB);

			SMSC_TRACE2(DBG_RX,"RCB %p retrieved from array element %d and zeroing that element\r\n", pRCB, Adapter->BulkInExpectedSeq);

			Adapter->BulkInCompletionArray[Adapter->BulkInExpectedSeq] = NULL;
			Adapter->BulkInExpectedSeq = (Adapter->BulkInExpectedSeq + 1) % Adapter->MaxRCBs;

			Adapter->nBulkInCompletedSuccessfully++;
			if (pRCB->ulSize) {
				SMSC_TRACE1(DBG_RX,"   Read %d bytes\r\n", pRCB->ulSize);
				Adapter->nBulkInBytesRead += pRCB->ulSize;
				if (Adapter->StopBulkInOnRxEmpty)
						MP_SET_FLAG(Adapter, fMP_POST_READS);

				if (!MP_TEST_FLAG(Adapter, fMP_IGNORE_READS))
						bIndicateReceive = TRUE;
			}
			else
			{
				SMSC_TRACE0(DBG_RX,"   ZLP !!!\r\n");

				if (Adapter->StopBulkInOnRxEmpty)
						MP_CLEAR_FLAG(Adapter, fMP_POST_READS); 		
			}

			if (bIndicateReceive)
			{
				
				InsertTailList(&Adapter->BulkInWaitList, &pRCB->List);

				NdisInterlockedIncrement(&Adapter->nWaitingRecv);
				ASSERT(Adapter->nWaitingRecv <= Adapter->MaxRCBs);

				SMSC_TRACE1(DBG_RX,"RCB %p removed from the busy and inserted into BulkInWaitList\r\n", pRCB);

				NdisReleaseSpinLock(&Adapter->BulkInLock);

				// Now process all the RCBs we've collected in the wait list
				NICProcessBulkInWaitList(Adapter);
			}
			else
			{
				// Put it in the Processed list.
				InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
				NdisInterlockedIncrement(&Adapter->nProcessedRecv);
				ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

				SMSC_TRACE1(DBG_RX,"RCB %p removed from it's current list (busy if in any) and inserted into BulkInProcessedList\r\n", pRCB);

				NdisReleaseSpinLock(&Adapter->BulkInLock);
				SMSC_TRACE1(DBG_RX,"pRCB->Ref 0x%x\r\n",pRCB->Ref);
				if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
				{
					NICFreeRCB(pRCB);
				}
				else
				{
					// This should never happen
					SMSC_TRACE0(DBG_ERROR,"NICReadRequestCompletion(place2) finds RCB with invalid Ref count!\r\n");
					SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
				}
			}

			NdisAcquireSpinLock(&Adapter->BulkInLock);

		} while(Adapter->BulkInCompletionArray[Adapter->BulkInExpectedSeq]);


	}

	NdisReleaseSpinLock(&Adapter->BulkInLock);
	
	NdisSetEvent(&Adapter->BulkInCompletionWorkItemReady);	
	MP_DEC_REF(Adapter);
	SMSC_TRACE0(DBG_RX,"<-- BulkInCompletionWorkItemCallBack\r\n");
	return STATUS_SUCCESS;
}





/*++

Routine Description:

	WorkItem used to proccess BulkInCompletionList
	
Arguments:

		WorkItem - Pointer to workitem
		Context -	pointer to the context structure (RCB)
	
Return Value:

	STATUS_SUCCESS
--*/


NTSTATUS
BulkInCompletionCallBack(
	PMP_ADAPTER Adapter
	)
{
	
	PRCB		   pRCB;
	BOOLEAN bIndicateReceive, bBulkInFailed = FALSE;
	ULONG	seq; 
	USB_ERROR	UsbErr;
	USB_TRANSFER hTransfer;
	BOOL dwErr=FALSE;
	int timeout=0;

	SMSC_TRACE0(DBG_RX,"--> BulkInCompletionWorkItemCallBack \r\n");

	if (Adapter->MPInitializeFail ||(!MP_TEST_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED)))
			return STATUS_SUCCESS;


	 NdisAcquireSpinLock(&Adapter->BulkInLock);
	while (!IsListEmpty(&Adapter->BulkInCompletionList) ) {

	   
		pRCB = (PRCB) RemoveHeadList(&Adapter->BulkInCompletionList);
		ASSERT(pRCB && (pRCB != (PRCB) &Adapter->BulkInCompletionList));// cannot be NULL and list cannot be empty
		NdisInterlockedDecrement(&Adapter->nCompletionRecv);
		ASSERT(Adapter->nCompletionRecv >= 0);
		NdisReleaseSpinLock(&Adapter->BulkInLock);
		
		seq= pRCB->seq;

		if (pRCB->ulCopied != 0)
			SMSC_TRACE1(DBG_ERROR,"BulkIn completion, pRCB->ulCopied = %d != 0,  \r\n",pRCB->ulCopied);
		ASSERT(pRCB->ulCopied == 0);

		//Sometimes the completion routine is called before the transfer complete. We need to sync. it here.
		//wait for Bulk In request to finish, so we can proccess a valid Bulk In transfer handler
		WaitForSingleObject(pRCB->SyncEvent, 100);	
		ResetEvent(pRCB->SyncEvent);

		if (pRCB->dwUsbErr == 0) {
			SMSC_TRACE1(DBG_ERROR,"ReadRequest has NULL Transfer,  hTransfer=0x%x \r\n", pRCB->dwUsbErr);
		}

		hTransfer=(USB_TRANSFER)pRCB->dwUsbErr;

		dwErr=GetTransferStatus(Adapter->UsbFuncs, hTransfer, &pRCB->dwTransferLength,&UsbErr);

		if (hTransfer){
			CloseTransferHandle(Adapter->UsbFuncs, hTransfer);
		}

		if( !(USB_NO_ERROR == UsbErr) || !dwErr || (Adapter->BulkIn.PipeTransferAborted) || !hTransfer)
		{
			bBulkInFailed = TRUE;
			Adapter->nBulkInCompletedWithError++;
			SMSC_TRACE5(DBG_ERROR,"ReadRequest failed dwErr = %x, dwUsbErr = %x, Aborted= %d, hTransfer=RF0x%x,RCB = %p \r\n",	\
										   dwErr, UsbErr, Adapter->BulkIn.PipeTransferAborted, hTransfer,pRCB);
		
			//
			// Clear the flags to prevent any more reads from being
			// posted to the target device.
			//
			MP_CLEAR_FLAG(Adapter, fMP_POST_READS);
			MP_SET_FLAG(Adapter, fMP_IGNORE_READS);

			// Handle USB Error recovery
			if ((UsbErr != USB_CANCELED_ERROR) && \
				(!Adapter->BulkIn.PipeTransferAborted) &&\
				(!MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS)) && \
				(!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) && \
				(!MP_TEST_FLAG(Adapter, fMP_DEVICE_NOT_ATTACHED)) )
			{
			 
				SMSC_TRACE2(DBG_ERROR,"NICReadRequestCompletion. USB Error !!! dwErr = %d, URB_Status = 0x%08lx\r\n",dwErr, UsbErr);

				// If device still present and workitem not already scheduled queue recovery.
				if (InterlockedExchange(&Adapter->IsUsbErrorRecoveryWorkItemQueued, TRUE) == FALSE)
				{
						Adapter->nUsbErrorRecoveryWorkItemScheduled++;
						MP_INC_REF(Adapter);
						NdisScheduleWorkItem(&Adapter->UsbErrorRecoveryWorkItem);
				}
			}

		}


		if (bBulkInFailed)
		{


			NdisAcquireSpinLock(&Adapter->BulkInLock);

			// Put it in the Processed list.
			InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedRecv);
			ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

			SMSC_TRACE1(DBG_RX,"RCB %p removed from it's current list (busy if in any) and inserted into BulkInProcessedList\r\n", pRCB);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
			{
				NICFreeRCB(pRCB);
			}
			else
			{
				// This should never happen
				SMSC_TRACE0(DBG_ERROR,"NICReadRequestCompletion(place1) finds RCB with invalid Ref count!\r\n");
				SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
			}
		 NdisAcquireSpinLock(&Adapter->BulkInLock);
			continue;
		}

		pRCB->ulSize = pRCB->dwTransferLength;
		
		NdisAcquireSpinLock(&Adapter->BulkInLock);

		// HOLD THIS LOCK UNTIL THE RCB IS INSERTED INTO THE NEXT APPPLICABLE
		// QUEUE. THIS IS SPECIALLY VITAL TO PREVENT GETTING OUT OF ORDER IN
		// THE WAIT LIST !!!

		ASSERT(Adapter->BulkInCompletionArray[seq] == NULL);
		Adapter->BulkInCompletionArray[seq] = pRCB;
		SMSC_TRACE2(DBG_RX,"RCB %p stored in array position %d\r\n", pRCB, seq);

		if (seq != Adapter->BulkInExpectedSeq)
		{
			SMSC_TRACE3(DBG_ERROR,"RCB %p out of order ! seq = %d, Expected seq = %d\r\n", pRCB, seq, Adapter->BulkInExpectedSeq);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
			{
				// For performance, instead of scheduling reads at the end of
				// every read completion, we will do it only when the number of
				// outstanding RCBs goes below Adapter->LowWatermarkRCBs.
//				if(((!Adapter->LowWatermarkRCBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkRCBs))
//					&& MP_TEST_FLAG(Adapter, fMP_POST_READS))
				if(MP_TEST_FLAG(Adapter, fMP_POST_READS))
				{
					NICPostReads(Adapter);
				}
			}
		 NdisAcquireSpinLock(&Adapter->BulkInLock);
			continue;
		}

		do
		{
			bIndicateReceive = FALSE;

			pRCB = Adapter->BulkInCompletionArray[Adapter->BulkInExpectedSeq];

			ASSERT(pRCB);

			SMSC_TRACE2(DBG_RX,"RCB %p retrieved from array element %d and zeroing that element\r\n", pRCB, Adapter->BulkInExpectedSeq);

			Adapter->BulkInCompletionArray[Adapter->BulkInExpectedSeq] = NULL;
			Adapter->BulkInExpectedSeq = (Adapter->BulkInExpectedSeq + 1) % Adapter->MaxRCBs;

			Adapter->nBulkInCompletedSuccessfully++;
			if (pRCB->ulSize) {
				SMSC_TRACE1(DBG_RX,"   Read %d bytes\r\n", pRCB->ulSize);
				Adapter->nBulkInBytesRead += pRCB->ulSize;
				if (Adapter->StopBulkInOnRxEmpty)
						MP_SET_FLAG(Adapter, fMP_POST_READS);

				if (!MP_TEST_FLAG(Adapter, fMP_IGNORE_READS))
						bIndicateReceive = TRUE;
			}
			else
			{
				SMSC_TRACE0(DBG_RX,"   ZLP !!!\r\n");

				if (Adapter->StopBulkInOnRxEmpty)
						MP_CLEAR_FLAG(Adapter, fMP_POST_READS); 		
			}

			if (bIndicateReceive)
			{
				
				InsertTailList(&Adapter->BulkInWaitList, &pRCB->List);

				NdisInterlockedIncrement(&Adapter->nWaitingRecv);
				ASSERT(Adapter->nWaitingRecv <= Adapter->MaxRCBs);

				SMSC_TRACE1(DBG_RX,"RCB %p removed from the busy and inserted into BulkInWaitList\r\n", pRCB);

				NdisReleaseSpinLock(&Adapter->BulkInLock);

				// Now process all the RCBs we've collected in the wait list
				NICProcessBulkInWaitList(Adapter);
			}
			else
			{
				// Put it in the Processed list.
				InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
				NdisInterlockedIncrement(&Adapter->nProcessedRecv);
				ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

				SMSC_TRACE1(DBG_RX,"RCB %p removed from it's current list (busy if in any) and inserted into BulkInProcessedList\r\n", pRCB);

				NdisReleaseSpinLock(&Adapter->BulkInLock);
				SMSC_TRACE1(DBG_RX,"pRCB->Ref 0x%x\r\n",pRCB->Ref);
				if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
				{
					NICFreeRCB(pRCB);
				}
				else
				{
					// This should never happen
					SMSC_TRACE0(DBG_ERROR,"NICReadRequestCompletion(place2) finds RCB with invalid Ref count!\r\n");
					SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
				}
			}

			NdisAcquireSpinLock(&Adapter->BulkInLock);

		} while(Adapter->BulkInCompletionArray[Adapter->BulkInExpectedSeq]);


	}

	NdisReleaseSpinLock(&Adapter->BulkInLock);
	
	SMSC_TRACE0(DBG_RX,"<-- BulkInCompletionWorkItemCallBack\r\n");
	return STATUS_SUCCESS;
}



/*++

Routine Description:

		Split out packets from the Bulk In RCB and indicate to NDIS.
	There will always be an integer number of packets in Bulk In RCB .

Arguments:

		pRCB - pointer to the RCB block

Return value:

	TRUE		All incoming RCB data has been processed and it's now in the processed list
	FALSE		Not all incoming RCB data has been processed and it's now in the wait list
--*/

BOOLEAN
NICProcessMultipleReceivePackets(
	IN PRCB 			pRCB,
	IN ULONG			BytesInUrb
	)
{
	ULONG			  PacketLength = 0, AlignCount = 0, BytesToCopy;
	PMP_ADAPTER 	  Adapter = pRCB->Adapter;
	PRCB			  SplitPktRCB = NULL;
	BOOLEAN 		  bIncomingRCBComplete = TRUE;
//	BOOLEAN 		  bVLANFiltered = FALSE;
	PUCHAR			  pDst;
	 ULONG             RxCmdA, RxCmdB;
	 

	SMSC_TRACE2(DBG_RX,"--> NICProcessMultipleReceivedPackets (RCB = %p, size = 0x%04x)\r\n", pRCB, BytesInUrb);

	ASSERT(BytesInUrb <= pRCB->ulSize);
	ASSERT(pRCB->ulCopied <= pRCB->ulSize);

	while (BytesInUrb)
	{
		
		ASSERT(BytesInUrb <= pRCB->ulSize);
		ASSERT(pRCB->ulCopied <= pRCB->ulSize);

		AlignCount = (4 - (PacketLength % 4)) % 4;	// Get number of bytes to align the Rx packet to dword
		SMSC_TRACE1(DBG_RX,"Rx pad AlignCount = %d\r\n", AlignCount);
		if (BytesInUrb < AlignCount)
		{
			// This should never happen since the URB is of size a multiple of 4 !!!
			ASSERT(FALSE);
			SMSC_TRACE3(DBG_ERROR,"BytesInUrb = %d <= %d = AlignCount. Not enough bytes to align till next packet, discarding left data in RCB %p\r\n", BytesInUrb, AlignCount, pRCB);
			break;
		}
		
		ASSERT((pRCB->ulCopied + AlignCount) <= pRCB->ulSize);

		pRCB->ulCopied += AlignCount;
		BytesInUrb -= AlignCount;

		// Since the Bulk IN size is a multiple of 4, and we just aligned ourselves to a multiple of 4
		// offset as well, at this point we have either:
		//			- a piece of a packet of size multiple of 4 bytes
		//			- a complete packet.
		//			- 0 bytes left here

		if (!BytesInUrb)
		{
			SMSC_TRACE0(DBG_ERROR,"BytesInUrb = 0, ethernet packet will start in next URB !\r\n");
			break;
		}

		// If we got here we have 8 or more bytes for sure
		ASSERT(BytesInUrb >=8);

		
		NdisAcquireSpinLock(&Adapter->RxPktLock);
		
		SplitPktRCB = (PRCB) RemoveHeadList(&Adapter->RxPktFreeList);
		
		if(SplitPktRCB == (PRCB) &Adapter->RxPktFreeList)
		{
			SMSC_TRACE2(DBG_RX,"No split packet RCB available to copy data into. RCB %p holding uncopied data at offset 0x%04x put at head of the Bulk In wait list\r\n", pRCB, pRCB->ulCopied);
			NdisReleaseSpinLock(&Adapter->RxPktLock); 
			bIncomingRCBComplete = FALSE;
			break;
		}
		
		// no need to initialize the list pointers since we are immediately storing in another queue
		
		SMSC_TRACE1(DBG_RX,"Retrieved RCB %p from RxPktFreeList and put it in RxPktBusyList\r\n", SplitPktRCB);
		
		ASSERT(SplitPktRCB->ulSize == 0);
		
		NdisInterlockedIncrement(&Adapter->nRxPktBusy);
		ASSERT(Adapter->nRxPktBusy <= Adapter->MaxRxPkts);
		
		InsertTailList(&Adapter->RxPktBusyList, &SplitPktRCB->List);
		NdisReleaseSpinLock(&Adapter->RxPktLock); 

		  // Keep a copy of Rx Command A & B
        RxCmdA = *((DWORD *)(pRCB->pData + pRCB->ulCopied));
        RxCmdB = *((DWORD *)(pRCB->pData + pRCB->ulCopied + 4));
        
        // Include the SW !
        BytesToCopy = ( *((DWORD *)(pRCB->pData + pRCB->ulCopied)) & RX_CMD_A_LEN_ ) + 8;   // Why +8 ? Add Status Words

        ASSERT(BytesToCopy <= NIC_BUFFER_SIZE);
		ASSERT(BytesToCopy <= BytesInUrb);
        
        if (BytesToCopy > BytesInUrb)
        {
        
            SMSC_TRACE2(DBG_ERROR,"BytesToCopy = %d > %d = BytesInUrb with BURST_CAP !=0 !!!\r\n", BytesToCopy, BytesInUrb);
        }
        

// OK, copy the data to the new RCB

        ASSERT((pRCB->ulCopied + BytesToCopy) <= pRCB->ulSize);

        ASSERT((SplitPktRCB->ulSize + BytesToCopy) <= NIC_BUFFER_SIZE);

        pDst = SplitPktRCB->pData;

		ASSERT (BytesToCopy <= NIC_BUFFER_SIZE);
		if (BytesToCopy > NIC_BUFFER_SIZE) {
			SMSC_TRACE2(DBG_ERROR,"RxCmdA = 0x%08x, RxCmdB = 0x%08x\r\n, i= %d",RxCmdA,RxCmdB);
			SMSC_TRACE5(DBG_ERROR,"BytesToCopy = %d > 2028, AlignCount= %d, pRCB->ulCopied =%d,pRCB->ulSize =%d,BytesInUrb= %d\r\n",BytesToCopy,AlignCount,pRCB->ulCopied,pRCB->ulSize,BytesInUrb);
			}

		if (Adapter->PriorityVlanTag != 0)
		   {
				SMSC_TRACE2(DBG_RX,"RxCmdA= 0x%x , RxCmdB= 0x%x\r\n", RxCmdA,RxCmdB);    		   
				if (RxCmdA & RX_CMD_A_FVTG_) // Vlan Tagged frame
				   {
					   // Packet has a VLAN tag !
					   NDIS_PACKET_8021Q_INFO VlanPriInfo;
					   UINT32 RxdVlanId;
					   
					   RxdVlanId = (UINT32)(RxCmdB & RX_CMD_B_VTAG_);
					   VlanPriInfo.Value = 0;
					   VlanPriInfo.TagHeader.VlanId = (RxdVlanId & 0xFFFUL);
					   VlanPriInfo.TagHeader.CanonicalFormatId = ((RxdVlanId & 0x1000UL)?1:0);
					   VlanPriInfo.TagHeader.UserPriority = ((RxdVlanId & 0xE000UL)>>13);
					   NDIS_PER_PACKET_INFO_FROM_PACKET(SplitPktRCB->Packet, Ieee8021QInfo) = VlanPriInfo.Value; 
				
				   }			
		   }
		   
		
		
		NdisMoveMemory( pDst,
						pRCB->pData + pRCB->ulCopied,
						BytesToCopy);

		SMSC_TRACE5(DBG_RX,"Copying %d bytes from %p in RCB %p to %p in RCB %p\r\n",
								BytesToCopy,
								pRCB->pData + pRCB->ulCopied,
								pRCB,
								pDst,
								SplitPktRCB);
	

		 // Set NDIS packet extension info according to RxCmdA
        if (Adapter->NicChecksumOffload.DoRcvIpChecksum ||
            Adapter->NicChecksumOffload.DoRcvTcpChecksum ||
            Adapter->NicChecksumOffload.DoRcvUdpChecksum)
        {
            PNDIS_PACKET_EXTENSION pPktExtInfo;
            PNDIS_TCP_IP_CHECKSUM_PACKET_INFO ChecksumPktInfo;

            // Make sure this is a IPv4 packet
            if (!(RxCmdA & RX_CMD_A_IPV_) && // IPv4
                 (RxCmdA & RX_CMD_A_PID_))   // IP, TCP, or UDP
            {
                pPktExtInfo = NDIS_PACKET_EXTENSION_FROM_PACKET(SplitPktRCB->Packet);
                if (pPktExtInfo == NULL)
                {
                    SMSC_TRACE0(DBG_RX,"NDIS_PACKET_EXTENSION_FROM_PACKET returns NULL in NICProcessMultipleReceivePackets()!!!\r\n");
                    goto Exit_MultiRxChecksumOffload;
                }
                ChecksumPktInfo = (PNDIS_TCP_IP_CHECKSUM_PACKET_INFO)&(pPktExtInfo->NdisPacketInfo[TcpIpChecksumPacketInfo]);
                ChecksumPktInfo->Value = 0UL;
                if (Adapter->NicChecksumOffload.DoRcvIpChecksum)
                {
                    // Set Rx Checksum flag in NDIS_PER_PACKET_INFO_FROM_PACKET
                    if (RxCmdA & RX_CMD_A_ICE_) // IP Checksum Error
                    {
                        ChecksumPktInfo->Receive.NdisPacketIpChecksumFailed = 1;
                    }
                    else
                    {
                        ChecksumPktInfo->Receive.NdisPacketIpChecksumSucceeded = 1;
                    }
                }

                if (RxCmdA & RX_CMD_A_TCE_) // TCP/UDP Checksum Error
                {
                    switch (RxCmdA & RX_CMD_A_PID_)
                    {
                        case RX_CMD_A_PID_TCP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvTcpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketTcpChecksumFailed = 1;
                            }
                            break;

                        case RX_CMD_A_PID_UDP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvUdpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketUdpChecksumFailed = 1;
                            }
                        default:    
                            break;
                    }
                }
                else
                {
                    switch (RxCmdA & RX_CMD_A_PID_)
                    {
                        case RX_CMD_A_PID_TCP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvTcpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketTcpChecksumSucceeded = 1;
                            }
                            break;

                        case RX_CMD_A_PID_UDP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvUdpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketUdpChecksumSucceeded = 1;
                            }
                        default:    
                            break;
                    }
                }
                
//                pPktExtInfo->NdisPacketInfo[TcpIpChecksumPacketInfo] = (PVOID)ChecksumPktInfo.Value;
            }
        }
       
Exit_MultiRxChecksumOffload:

        // Point to the next packet (or EOP) inside of the original RCB
        pRCB->ulCopied += BytesToCopy;
        ASSERT (BytesInUrb >= BytesToCopy);
        BytesInUrb -= BytesToCopy;      
        ASSERT (BytesInUrb <= pRCB->ulSize);

        SplitPktRCB->ulSize += BytesToCopy;

//        if (bVLANFiltered == FALSE)
//        {
            NICIndicateReceiveSplitPacket(SplitPktRCB, &PacketLength);
//        }

//		else
//		{
	/*
			PacketLength = SplitPktRCB->ulSize - 8UL;	// Need for alignment calculations at the top

			// Packet will be dropped, put SplitPktRCB back in free list.
			NdisAcquireSpinLock(&Adapter->RxPktLock);

			RemoveEntryList(&SplitPktRCB->List);
			NdisInterlockedDecrement(&Adapter->nRxPktBusy);
			ASSERT(Adapter->nRxPktBusy >= 0);

			SplitPktRCB->ulSize = 0;

			InsertTailList(&Adapter->RxPktFreeList, &SplitPktRCB->List);

			SMSC_TRACE1(DBG_RX,"RCB %p removed from RxPktBusy list inserted into RxPktFreeList\r\n", SplitPktRCB);

			NdisReleaseSpinLock(&Adapter->RxPktLock); 

			bVLANFiltered = FALSE;
*/
//		}
        
    }

    // Insert the Bulk in RCB in the proper list
    NdisAcquireSpinLock(&Adapter->BulkInLock);
    if (bIncomingRCBComplete)
    {
        // Put it in the Processed list.
        InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
        NdisInterlockedIncrement(&Adapter->nProcessedRecv);
    }
    else
    {
        // Increment the reference to compensate the decrement that 
        // will be done as soon as we return from this function
        NdisInterlockedIncrement(&pRCB->Ref);

        // Put it in the head of the Wait list.
        InsertHeadList(&Adapter->BulkInWaitList, &pRCB->List);
        NdisInterlockedIncrement(&Adapter->nWaitingRecv);
    }
    NdisReleaseSpinLock(&Adapter->BulkInLock);

    
    return bIncomingRCBComplete;
}


VOID
NICProcessSingleReceivePacket(
	IN PRCB 			pRCB
    )
/*++

Routine Description:

    Initialize the packet to describe the received data and
    indicate to NDIS.
        
Arguments:

    pRCB - pointer to the RCB block    
    BytesToIndicate - number of bytes to indicate

Return value:

    VOID
--*/
{
    ULONG           PacketLength, FrameLen;
    PNDIS_BUFFER    CurrentBuffer = NULL;
    PMP_ADAPTER     Adapter = pRCB->Adapter;
    PNDIS_PACKET    Packet = pRCB->Packet;
    NDIS_STATUS     Status =  NDIS_STATUS_SUCCESS;
//    BOOLEAN         bVLANFiltered = FALSE;
    ULONG           RxCmdA, RxCmdB;
    

    SMSC_TRACE1(DBG_RX,"--->NICProcessSingleReceivePacket, RCB = %p\r\n", pRCB);


	

    // Keep a copy of Rx Command A & B
    RxCmdA = *((DWORD *)pRCB->pData);
    RxCmdB = *((DWORD *)(pRCB->pData + 4));

    FrameLen = ( *((DWORD *)pRCB->pData) & RX_CMD_A_LEN_ );

//  ASSERT( (FrameLen  == (BytesToIndicate - 8UL)) && (FrameLen <= (NIC_BUFFER_SIZE - 4UL) ) );

		
    if (Adapter->PriorityVlanTag != 0)
    {
		SMSC_TRACE2(DBG_RX,"RxCmdA= 0x%x,RxCmdB= 0x%x  \r\n", RxCmdA,RxCmdB); 

         if (RxCmdA & RX_CMD_A_FVTG_) // Vlan Tagged frame
            {
                // Packet has a VLAN tag !
                NDIS_PACKET_8021Q_INFO VlanPriInfo;
                UINT32 RxdVlanId;
       
                RxdVlanId = (UINT32)(RxCmdB & RX_CMD_B_VTAG_);                
                VlanPriInfo.Value = 0;
                VlanPriInfo.TagHeader.VlanId = (RxdVlanId & 0xFFFUL);
                VlanPriInfo.TagHeader.CanonicalFormatId = ((RxdVlanId & 0x1000UL)?1:0);
                VlanPriInfo.TagHeader.UserPriority = ((RxdVlanId & 0xE000UL)>>13);
                NDIS_PER_PACKET_INFO_FROM_PACKET(Packet, Ieee8021QInfo) = VlanPriInfo.Value; 

				
            }
    }

	
    // Use the frame length reported in the Rx Status word (which already substracts
    // the 8 byte length of the Sts wds itself). Just substract 4 more to remove the FCS 
    // (ethernet CRC) at the end of the frame. 2 more for 16-bits of alignment padding in front of all receiving packets
    NdisAdjustBufferLength(pRCB->Buffer, FrameLen - 4UL - 2UL);

    //
    // Prepare the recv packet
    //
    NdisReinitializePacket(Packet);

    *((PRCB *)Packet->MiniportReservedEx) = pRCB;       //Save the RCB pointer for later.

    //
    // Chain the RCB buffers to the packet. Note that when the Ndis buffer
    // was allocated (in NICAllocRecvResources) it was made pointing to 
    // offset 8 (skipping Rx Sts wds) in the pRCB->pData data buffer.
    //
    NdisChainBufferAtBack(Packet, pRCB->Buffer);    

    NdisQueryPacket(Packet, NULL, NULL, &CurrentBuffer, (PUINT) &PacketLength);

    ASSERT(CurrentBuffer == pRCB->Buffer);

    
    //RK could move up to discard earlier ?
    if(NICPacketPassesRxFilter(Adapter, (PDWORD)pRCB->pData))
    {
        SMSC_TRACE2(DBG_RX,"Indicating packet = %p, Packet Length = %d\n", 
                        pRCB->Packet, PacketLength);

        // Set NDIS packet extension info according to RxCmdA
        if (Adapter->NicChecksumOffload.DoRcvIpChecksum ||
            Adapter->NicChecksumOffload.DoRcvTcpChecksum ||
            Adapter->NicChecksumOffload.DoRcvUdpChecksum)
        {
            PNDIS_PACKET_EXTENSION pPktExtInfo;
            PNDIS_TCP_IP_CHECKSUM_PACKET_INFO ChecksumPktInfo;

            // Make sure this is a IPv4 packet
            if (!(RxCmdA & RX_CMD_A_IPV_) && // IPv4
                 (RxCmdA & RX_CMD_A_PID_))   // IP, TCP, or UDP
            {
                pPktExtInfo = NDIS_PACKET_EXTENSION_FROM_PACKET(pRCB->Packet);
                if (pPktExtInfo == NULL)
                {
                    SMSC_TRACE0(DBG_RX,"NDIS_PACKET_EXTENSION_FROM_PACKET returns NULL in NICProcessSingleReceivePacket()!!!\r\n");
                    goto Exit_SingleRxChecksumOffload;
                }
                ChecksumPktInfo = (PNDIS_TCP_IP_CHECKSUM_PACKET_INFO)&(pPktExtInfo->NdisPacketInfo[TcpIpChecksumPacketInfo]);
                ChecksumPktInfo->Value = 0UL;
                if (Adapter->NicChecksumOffload.DoRcvIpChecksum)
                {
                    // Set Rx Checksum flag in NDIS_PER_PACKET_INFO_FROM_PACKET
                    if (RxCmdA & RX_CMD_A_ICE_) // IP Checksum Error
                    {
                        ChecksumPktInfo->Receive.NdisPacketIpChecksumFailed = 1;
                    }
                    else
                    {
                        ChecksumPktInfo->Receive.NdisPacketIpChecksumSucceeded = 1;
                    }
                }

                if (RxCmdA & RX_CMD_A_TCE_) // TCP/UDP Checksum Error
                {
                    switch (RxCmdA & RX_CMD_A_PID_)
                    {
                        case RX_CMD_A_PID_TCP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvTcpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketTcpChecksumFailed = 1;
                            }
                            break;

                        case RX_CMD_A_PID_UDP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvUdpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketUdpChecksumFailed = 1;
                            }
                        default:    
                            break;
                    }
                }
                else
                {
                    switch (RxCmdA & RX_CMD_A_PID_)
                    {
                        case RX_CMD_A_PID_TCP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvTcpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketTcpChecksumSucceeded = 1;
                            }
                            break;

                        case RX_CMD_A_PID_UDP_IP_:
                            if (Adapter->NicChecksumOffload.DoRcvUdpChecksum)
                            {
                                ChecksumPktInfo->Receive.NdisPacketUdpChecksumSucceeded = 1;
                            }
                        default:    
                            break;
                    }
                }
                
//                pPktExtInfo->NdisPacketInfo[TcpIpChecksumPacketInfo] = (PVOID)ChecksumPktInfo.Value;
            }
        }
        
Exit_SingleRxChecksumOffload:

        NdisInterlockedIncrement(&pRCB->Ref);

        Adapter->nPacketsIndicatedOnRx++;

        // Put it in the Processed list BEFORE the indication (NDIS can call  
        // MPReturnPacket synchronously before returning from NdisMIndicateReceivePacket.

        NdisAcquireSpinLock(&Adapter->BulkInLock);

        SMSC_TRACE1(DBG_RX,"RCB %p inserted into BulkInProcessedList\r\n", pRCB);

        InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
        NdisInterlockedIncrement(&Adapter->nProcessedRecv);
        ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

        // If we are running out of RCBs tell NDIS to give it back immediately
        if((!Adapter->LowWatermarkRCBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkRCBs))
        {
            Status = NDIS_STATUS_RESOURCES;
        }
        NdisReleaseSpinLock(&Adapter->BulkInLock);

        NDIS_SET_PACKET_STATUS(pRCB->Packet, Status);
        SMSC_TRACE4(DBG_RX,"Indicating RCB/packet %p/%p (size %d) to Ndis with status 0x%08lx\r\n", pRCB, pRCB->Packet, PacketLength, Status);

        NdisMIndicateReceivePacket(Adapter->AdapterHandle, &pRCB->Packet, 1);

        if (Status == NDIS_STATUS_RESOURCES)
        {
            MPReturnPacket((NDIS_HANDLE) Adapter, pRCB->Packet);
        }

        goto Exit_NICProcessSingleReceivePacket;

    }
    else
    {
        SMSC_TRACE2(DBG_RX,"Rx Packet (Single) was filtered. Packet = %p,Packet Length = %d\r\n", pRCB->Packet, PacketLength);        
    }            

    // Put it in the Processed list.
    NdisAcquireSpinLock(&Adapter->BulkInLock);

    SMSC_TRACE1(DBG_RX,"RCB %p inserted into BulkInProcessedList\r\n", pRCB);
    
    InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
    NdisInterlockedIncrement(&Adapter->nProcessedRecv);
    ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

    NdisReleaseSpinLock(&Adapter->BulkInLock);

Exit_NICProcessSingleReceivePacket:
	 SMSC_TRACE0(DBG_RX,"<---NICProcessSingleReceivePacket\r\n");

}

VOID
NICIndicateReceiveSplitPacket(
    IN PRCB             pRCB,
    PULONG              pPacketLength
    )
/*++

Routine Description:

    Initialize the packet to describe the received data and
    indicate to NDIS. If the packet is bad don't indicate it.
        
Arguments:

    pRCB - pointer to the RCB block    

Return value:

    VOID
--*/

{
    PMP_ADAPTER       Adapter = pRCB->Adapter;
    PNDIS_PACKET      Packet = NULL;
    PNDIS_BUFFER      CurrentBuffer = NULL;
    
    NDIS_STATUS       Status =  NDIS_STATUS_SUCCESS;

	SMSC_TRACE0(DBG_RX,"---> NICIndicateReceiveSplitPacket\r\n");

    // Prepare the actual packet and indicate it !
    // do not pass Status Words and FCS up !
    // Deduct 2 more bytes for 16-bits of alignment padding in front of all receiving packets
    NdisAdjustBufferLength(pRCB->Buffer, pRCB->ulSize - 12UL - 2UL); 

    //
    // Prepare the recv packet
    //
    Packet = pRCB->Packet;
    NdisReinitializePacket(Packet);

    *((PRCB *)Packet->MiniportReservedEx) = pRCB;       //Save the RCB pointer for later.

    //
    // Chain the RCB buffers to the packet. Note that when the Ndis buffer
    // was allocated (in NICAllocRecvResources) it was made pointing to 
    // offset 8 (skipping Rx Sts words) in the pRCB->pData data buffer.
    //
    NdisChainBufferAtBack(Packet, pRCB->Buffer);    

    NdisQueryPacket(Packet, NULL, NULL, &CurrentBuffer, (PUINT) pPacketLength);

    // Adjust 2 bytes alignment
    *pPacketLength += 2UL;
    
    ASSERT(CurrentBuffer == pRCB->Buffer);

    
    //RK could move up to discard earlier ?
    if(NICPacketPassesRxFilter(Adapter, (PDWORD)pRCB->pData))
    {
        NdisInterlockedIncrement(&pRCB->Ref);

        Adapter->nPacketsIndicatedOnRx++;

        // If we are running out of split packets tell NDIS to give it back immediately
        NdisAcquireSpinLock(&Adapter->RxPktLock);
        if(Adapter->nRxPktBusy == Adapter->MaxRxPkts)
        {
            Status = NDIS_STATUS_RESOURCES;
        }
        NDIS_SET_PACKET_STATUS(pRCB->Packet, Status);
        NdisReleaseSpinLock(&Adapter->RxPktLock); 

        SMSC_TRACE4(DBG_RX,"Indicating RCB/packet %p/%p (size %d) to Ndis with status 0x%08lx\n", 
                                pRCB, 
                                pRCB->Packet, 
                                *pPacketLength, 
                                Status);

        // The code is structured to do simple packet indications. If desired 
        // to implement multipacket indications (maybe indicate all packets 
        // from the same RCB at the same time) it will require some code 
        // restructuring which is do-able but not required at this time 
        // (Rx performance for this driver significantly higher than any
        // competiro).
        NdisMIndicateReceivePacket(Adapter->AdapterHandle, &pRCB->Packet, 1);

        if (Status == NDIS_STATUS_RESOURCES)
        {
            MPReturnPacket((NDIS_HANDLE) Adapter, pRCB->Packet);
        }

        goto Exit_NICIndicateReceiveSplitPacket;
    }
    else 
    {
        SMSC_TRACE2(DBG_RX,"Rx Packet (Multi) was filtered. Packet = %p,Packet Length = %d\n",
             pRCB->Packet, *pPacketLength);        
    }

    // Packet will be dropped, put RCB back in free list.
    NdisAcquireSpinLock(&Adapter->RxPktLock);

    RemoveEntryList(&pRCB->List);
    NdisInterlockedDecrement(&Adapter->nRxPktBusy);
    ASSERT(Adapter->nRxPktBusy >= 0);

    pRCB->ulSize = 0;

    InsertTailList(&Adapter->RxPktFreeList, &pRCB->List);

    SMSC_TRACE1(DBG_RX,"RCB %p removed from RxPktBusy list inserted into RxPktFreeList\r\n", pRCB);

    NdisReleaseSpinLock(&Adapter->RxPktLock); 

Exit_NICIndicateReceiveSplitPacket:
    	SMSC_TRACE0(DBG_RX,"<--- NICIndicateReceiveSplitPacket\r\n");
}


NIC_STATUS
NICStopAndFlushRxPath(
    PMP_ADAPTER Adapter
    )
/*++

Routine Description:

    This function stops all activity on the Rx Path, both at the USB and Mac
    and then flushes the Receive Fifo.
    
Arguments:

    Adapter    - pointer to the MP_ADAPTER structure
    
Return value:

    VOID
--*/
{ 
    NIC_STATUS Status = NIC_STATUS_FAILURE;

     // Since it can be called from MP_Halt when MP_Initialize fails
    if(!MP_TEST_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED))
    {
        goto Exit_NICStopAndFlushRxPath;
    }

    Status = NICStopRxPath(Adapter);
    Status |= NICFlushRxFifo(Adapter);

Exit_NICStopAndFlushRxPath:
 
    return (Status);
}



NIC_STATUS
NICFlushRxFifo(
    PMP_ADAPTER Adapter
    )
/*++

Routine Description:

    This function stops Rx FIFO then flush Rx FIFO and UTX
    
Arguments:

    Adapter    - pointer to the MP_ADAPTER structure
    
Return value:

    VOID
--*/
{ 
    DWORD Value32, Count = 0;
    NIC_STATUS Status;

   
    // Disable Rx FIFO and clear FCT Rx Disabled Flag if not already
    Status = NICReadRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
    Value32= ((Value32 & (~ETH_FCT_RX_CTL_EN_)) | ETH_FCT_RX_CTL_DIS_);
    Status |= NICWriteRegister(Adapter, ETH_FCT_RX_CTL, &Value32 );

    // Let's wait for 3 msec
    do 
    {
        NdisMSleep(500);
        Status |= NICReadRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
    }
    while ( (++Count<6) && ((Value32 & ETH_FCT_RX_CTL_DIS_) == 0) );

    if (!(Value32 & ETH_FCT_RX_CTL_DIS_))
    {
        SMSC_TRACE1(DBG_ERROR,"ETH_FCT_RX_CTL_DIS_ is not yet set after 3msec in NICFlushRxFifo!!!, ETH_FCT_RX_CTL=0x%08lX\r\n",Value32);
        Status = NIC_STATUS_FAILURE;
        goto Exit_NICFlushRxFifo;
    }
    
    // Reset Rx FIFO and clear FCT Rx Disabled
    Value32 |= ETH_FCT_RX_CTL_RST_ | ETH_FCT_RX_CTL_DIS_;
    Status |= NICWriteRegister(Adapter, ETH_FCT_RX_CTL, &Value32 );

    // Confirm Rx FIFO is flushed
    Status |= NICReadRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
    if (Value32 & ETH_FCT_RX_CTL_RST_)
    {
        SMSC_TRACE1(DBG_ERROR,"ETH_FCT_RX_CTL_RST_ is still set after 125usec in NICFlushRxFifo!!!, ETH_FCT_RX_CTL=0x%08lX\r\n", Value32);
        Status = NIC_STATUS_FAILURE;
    }

Exit_NICFlushRxFifo:
    // Re-enable Rx FIFO before leave the routine
    Value32 |= ETH_FCT_RX_CTL_EN_;
    Status |= NICWriteRegister(Adapter, ETH_FCT_RX_CTL, &Value32);
    return (Status);
}




/*++

Routine Description:

	This function stops all activity on the Rx Path, both at the USB and Mac

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return value:

	VOID
--*/

NIC_STATUS
NICStopRxPath(
	PMP_ADAPTER Adapter
	)
{
	DWORD Value32,Count=0;
	NIC_STATUS Status = NIC_STATUS_FAILURE;

   SMSC_TRACE0(DBG_RX,"--> NICStopRxPath\r\n");

   if(!MP_TEST_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED)){
		goto Exit_NICStopRxPath;
	}

	 // Disable the receiver at the Mac and clear MAC Rx Disabled Flag if not already
    Status = NICReadRegister(Adapter, ETH_MAC_RX, &Value32);
    Value32= ((Value32 & (~ETH_MAC_RX_RXEN_)) | ETH_MAC_RX_RXD_);
    Status |= NICWriteRegister(Adapter, ETH_MAC_RX, &Value32);

    // The Rx Stop bit should assert as soon as the packet "in flight" makes 
    // it into the Mac, worst case is 10 Mbps HD. This will be ~2ms tops
    // Let's wait for 3 msec
    do 
    {
        Status |= NICReadRegister(Adapter, ETH_MAC_RX, &Value32);
    }
    while ( (++Count<6) && ((Value32 & ETH_MAC_RX_RXD_) == 0) );

    Status |= NICStopBulkIn(Adapter);

	// If operating in multipacket mode ...

	if (!Adapter->RxOnePktperRcb)
	{
		NICCollectSplitRxPackets(Adapter);
	}

Exit_NICStopRxPath:

	SMSC_TRACE0(DBG_RX,"<-- NICStopRxPath\r\n");
	return Status;
}

/*++

Routine Description:

	This function stops all further USB bulk in submissions, then tries to
	cancel all the outstanding BulkIn transaction if it is not already completed, waits
	for their completion and frees the RCB block.

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return value:

	NIC_STATUS_XXX
--*/

NIC_STATUS
NICStopBulkIn(
	PMP_ADAPTER Adapter
	)
{
	BOOLEAN 		  bDone=TRUE;
	LONG			  nWaitCount = 0;
	int 			  i;
	PRCB			  pRCB;
	NIC_STATUS Status = NIC_STATUS_SUCCESS;
	

	SMSC_TRACE0(DBG_SHUT_DOWN,"--> NICStopBulkIn\r\n");

	if(!MP_TEST_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED)){
		goto Exit_NICStopBulkIn;
	}
	
	MP_CLEAR_FLAG(Adapter, fMP_POST_READS);
	MP_SET_FLAG(Adapter, fMP_IGNORE_READS);

	
	if (!MP_TEST_FLAG(Adapter, fMP_SURPRISE_REMOVED)) {
		// Try to stop them before the completion in one shot.	
		Adapter->BulkIn.PipeTransferAborted=TRUE;
		if ( ! Adapter->UsbFuncs->lpAbortPipeTransfers(Adapter->BulkIn.hPipe,USB_NO_WAIT))
		{
			SMSC_TRACE0(DBG_ERROR,"Bulk In Pipe Abort failed !!!\r\n");
					
		}
	}

	// Wait for all bulk in completions and RCBs to make it back to the free queue
	while(TRUE)
	{
		bDone = TRUE;

		// If we have RCBs waiting in the out of order array release them to the
		// free queue
		
		for (i=0; i<Adapter->MaxRCBs; i++)
		{
			NdisAcquireSpinLock(&Adapter->BulkInLock);
			pRCB = Adapter->BulkInCompletionArray[i];
			if (pRCB)
			{
				Adapter->BulkInCompletionArray[i] = NULL;

				
				InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
				NdisInterlockedIncrement(&Adapter->nProcessedRecv);
				ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);
				SMSC_TRACE1(DBG_RX,"RCB %p inserted into BulkInProcessedList\r\n", pRCB);

				NdisReleaseSpinLock(&Adapter->BulkInLock);

				// Decrement reference and free RCB if we can.
				if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
				{
					NICFreeRCB(pRCB);
				}
				else
				{
					// This should never happen
					SMSC_TRACE0(DBG_ERROR,"NICStopBulkIn finds RCB in the BulkInCompletionArray with invalid Ref count!\r\n");
					SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
				}
			}
			else
			{
				NdisReleaseSpinLock(&Adapter->BulkInLock);
			}
		}
		

		// If we have RCBs in the waiting list release them to the
		// free queue		
		NdisAcquireSpinLock(&Adapter->BulkInLock);

		while(!IsListEmpty(&Adapter->BulkInWaitList))
		{
			pRCB = (PRCB) RemoveHeadList(&Adapter->BulkInWaitList);
			ASSERT(pRCB && (pRCB != (PRCB) &Adapter->BulkInWaitList));// cannot be NULL and list cannot be empty

			NdisInterlockedDecrement(&Adapter->nWaitingRecv);
			ASSERT(Adapter->nWaitingRecv >= 0);


			InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedRecv);
			ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);
			SMSC_TRACE1(DBG_RX,"RCB %p inserted into BulkInProcessedList\r\n", pRCB);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			// Decrement reference and free RCB if we can.
			if (NdisInterlockedDecrement(&pRCB->Ref) == 0)
			{
				NICFreeRCB(pRCB);
			}
			
			ASSERT(&pRCB->Ref >= 0);

			NdisAcquireSpinLock(&Adapter->BulkInLock);
		}

		 NdisReleaseSpinLock(&Adapter->BulkInLock);

		//
		// Are all the bulk in transactions completed ?
		//
		if(Adapter->nBusyRecv + Adapter->nWaitingRecv + Adapter->nProcessedRecv + Adapter->nCompletionRecv)
		{
			SMSC_TRACE1(DBG_RX,"nBusyRecv = %d\r\n", Adapter->nBusyRecv);
			SMSC_TRACE1(DBG_RX,"nWaitingRecv = %d\r\n", Adapter->nWaitingRecv);
			SMSC_TRACE1(DBG_RX,"nProcessedRecv = %d\r\n", Adapter->nProcessedRecv);
			SMSC_TRACE1(DBG_RX,"nCompletionRecv = %d\r\n", Adapter->nCompletionRecv);
			bDone = FALSE;
		}
	   

		if(bDone)
		{
			break;
		}

		// Need to wait 120 seconds in case single packet RCBs in Processed list not returned
		// by Ndis on purpose (like in Ndistest 6.0's 2c_holdpackets)
		if((++nWaitCount % NIC_10_SEC_TIMEOUT) == 0)
		{
			SMSC_TRACE1(DBG_ERROR,"Bulk In Cancel timed out after %d seconds !!! COMMITING LEAK !!!\r\n", NIC_1_SEC_TIMEOUT / 1000);
			SMSC_TRACE1(DBG_ERROR,"   nBusyRecv = %d\r\n", Adapter->nBusyRecv);
			SMSC_TRACE1(DBG_ERROR,"   nWaitingRecv = %d\r\n", Adapter->nWaitingRecv);
			SMSC_TRACE1(DBG_ERROR,"   nProcessedRecv = %d\r\n", Adapter->nProcessedRecv);
			SMSC_TRACE1(DBG_ERROR,"nCompletionRecv = %d\r\n", Adapter->nCompletionRecv);
			Status = NIC_STATUS_FAILURE;
			ASSERT(FALSE);
			break;
		}

		NdisMSleep(NDIS_SLEEP_1_MS);
	}

Exit_NICStopBulkIn:   
	SMSC_TRACE0(DBG_SHUT_DOWN,"<-- NICStopBulkIn\r\n");
	return Status;
}

/*++

Routine Description:

	This function waits for all split Rx packets which were indicated
	to Ndis to be returned

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return value:

	VOID
--*/
VOID
NICCollectSplitRxPackets(
	PMP_ADAPTER Adapter
	)
{
	BOOLEAN 		  bDone=TRUE;
	LONG			  nWaitCount = 0;

	SMSC_TRACE0(DBG_RX,"--> NICCollectSplitRxPackets\r\n");

	// Wait for all split Rx packets to be completed and RCBs to make
	// it back to the free queue
	while(TRUE)
	{
		bDone = TRUE;

		//
		// Are all the Rx split packets returned ?
		//
		if(Adapter->nRxPktBusy)
		{
			SMSC_TRACE1(DBG_RX,"nRxPktBusy = %d\r\n", Adapter->nRxPktBusy);
			bDone = FALSE;
		}

		if(bDone)
		{
			break;
		}

		// Need to wait 120 seconds in case split RCBs in Processed list not returned
		// by Ndis on purpose (like in Ndistest 6.0's 2c_holdpackets)
		if((++nWaitCount % NIC_2_MIN_TIMEOUT) == 0)
		{
			SMSC_TRACE1(DBG_ERROR,"Split Packets return wait timed out after %d seconds !!!\r\n", NIC_2_MIN_TIMEOUT/1000);
			SMSC_TRACE1(DBG_ERROR,"COMMITING LEAK !!! Adapter->nRxPktBusy = %d packets !!!\r\n", Adapter->nRxPktBusy);
			ASSERT(FALSE);
			break;
		}


		NdisMSleep(NDIS_SLEEP_1_MS);
	}

	SMSC_TRACE0(DBG_RX,"<-- NICCollectSplitRxPackets\r\n");
}

/*++

Routine Description:

	This function tries to cancel all the outstanding Bulk In request if it is not
	already completed and frees the RCB block.

Arguments:

	Adapter    - pointer to the MP_ADAPTER structure

Return value:

	VOID
--*/
VOID
NICFreeBusyRecvPackets(
	PMP_ADAPTER Adapter
	)
{
	PRCB				pRCB = NULL;

	SMSC_TRACE0(DBG_RX,"--> NICFreeBusyRecvPackets\r\n");

	if(!MP_TEST_FLAG(Adapter, fMP_RECV_SIDE_RESOURCE_ALLOCATED)){
		return;
	}

	NdisAcquireSpinLock(&Adapter->BulkInLock);

	while(!IsListEmpty(&Adapter->BulkInBusyList))
	{
		pRCB = (PRCB) RemoveHeadList(&Adapter->BulkInBusyList);

		ASSERT(pRCB && (pRCB != (PRCB) &Adapter->BulkInBusyList));// cannot be NULL and list cannot be empty

		// Make sure if we use RemoveEntryList on this RCB later we don't corrupt the queue.
		NdisInitializeListHead(&pRCB->List);

		// Do not decrement nBusyRecv until the actual completion (either
		// real completion callback function or cancellation below)

		NdisReleaseSpinLock(&Adapter->BulkInLock);

 
		if(NdisInterlockedDecrement(&pRCB->Ref) == 0)
		{
			// Put it in the Processed list since NICFreeRCB expects it there.
			NdisAcquireSpinLock(&Adapter->BulkInLock);

			NdisInterlockedDecrement(&Adapter->nBusyRecv);
			ASSERT(Adapter->nBusyRecv >= 0);


			InsertTailList(&Adapter->BulkInProcessedList, &pRCB->List);
			NdisInterlockedIncrement(&Adapter->nProcessedRecv);
			ASSERT(Adapter->nProcessedRecv <= Adapter->MaxRCBs);

			SMSC_TRACE1(DBG_RX,"RCB %p inserted into BulkInProcessedList\r\n", pRCB);

			NdisReleaseSpinLock(&Adapter->BulkInLock);

			NICFreeRCB(pRCB);
		}
		else
		{
			// This should never happen
			SMSC_TRACE0(DBG_ERROR,"NICFreeBusyRecvPackets finds RCB with invalid Ref count!\r\n");
			SMSC_TRACE0(DBG_ERROR,"Only we have the right to free RCB\r\n");
		}

		NdisAcquireSpinLock(&Adapter->BulkInLock);
	}

	NdisReleaseSpinLock(&Adapter->BulkInLock);

	SMSC_TRACE0(DBG_RX,"<-- NICFreeBusyRecvPackets\r\n");

	return ;
}

/*++

Routine Description:

		This routine reinitializes the RCB block and puts it back
		into the BulkInFreeList (Bulk In RCB) or RxPktFreeList (Split pkt RCB)
		for reuse. On entry, it is assumed that if it is a Bulk In RCB it was
	in the Processed list, if it is a split packet RCB in the RxPktBusy
	list

Arguments:

	pRCB	  - pointer to RCB block

Return Value:

	VOID

--*/
VOID
NICFreeRCB(
	IN PRCB pRCB)
{
	PMP_ADAPTER Adapter = pRCB->Adapter;
	
	SMSC_TRACE1(DBG_RX,"--> NICFreeRCB %p\r\n", pRCB);

	ASSERT(!pRCB->Buffer->Next); // should be NULL
	ASSERT(!pRCB->Ref); // should be 0
	ASSERT(pRCB->Adapter); // shouldn't be NULL

	//
	// Re adjust the length to the original size
	//
//	 NdisAdjustBufferLength(pRCB->Buffer, (NIC_BUFFER_SIZE-10));
	
	// This could be a Bulk In RCB or a split packet RCB, their processing
	// is slightly different, so check first
	if (pRCB->BulkInRCB)
	{
		 NdisAdjustBufferLength(pRCB->Buffer, (Adapter->RxBurstMax - 10UL));
		// This is a Bulk in packet RCB
		pRCB->ulCopied = 0;

		//
		// Remove from the Processed List and insert the RCB back
		// in the Recv free list
		//
		NdisAcquireSpinLock(&Adapter->BulkInLock);

		RemoveEntryList(&pRCB->List);
		NdisInterlockedDecrement(&Adapter->nProcessedRecv);
		ASSERT(Adapter->nProcessedRecv >= 0);
		SMSC_TRACE1(DBG_RX,"RCB %p removed from the Processed List and inserted into BulkInFreeList\r\n", pRCB);
		InsertTailList(&Adapter->BulkInFreeList, &pRCB->List);

		NdisReleaseSpinLock(&Adapter->BulkInLock);

		if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
		{
			// For performance, instead of scheduling reads at the end of
			// every read completion, we will do it only when the number of
			// outstanding RCBs goes below Adapter->LowWatermarkRCBs.
			if(((!Adapter->LowWatermarkRCBs) || (Adapter->nBusyRecv <= Adapter->LowWatermarkRCBs))
					&& MP_TEST_FLAG(Adapter, fMP_POST_READS))
			{
				#ifdef USE_WORK_ITEM_RESUBMIT_RCB
				
					if (InterlockedExchange(&Adapter->IsReadWorkItemQueued, TRUE) == FALSE)
					{
						SMSC_TRACE0(DBG_RX,"NICFreeRCB, schedule ReadWorkItem\r\n");
						Adapter->nReadWorkItemScheduled++;
						MP_INC_REF(Adapter);
						NdisScheduleWorkItem(&Adapter->ReadWorkItem);
					}
					
				#else 
					if (Adapter->SubmitRCBFail) {
						if (InterlockedExchange(&Adapter->IsReadWorkItemQueued, TRUE) == FALSE)
						{
							SMSC_TRACE0(DBG_RX,"NICFreeRCB, schedule ReadWorkItem\r\n");
							Adapter->nReadWorkItemScheduled++;
							MP_INC_REF(Adapter);
							NdisScheduleWorkItem(&Adapter->ReadWorkItem);
						}
					} else {
							NICPostReads(Adapter);
						}
				#endif
			}
		}
	}
	else
	{
		// This is a split packet RCB

		//
		// Remove from the RxPkt busy list and insert the
		// RCB back in the RxPkt free list
		//
		 NdisAdjustBufferLength(pRCB->Buffer, (NIC_BUFFER_SIZE-10));
		NdisAcquireSpinLock(&Adapter->RxPktLock);

		RemoveEntryList(&pRCB->List);
		NdisInterlockedDecrement(&Adapter->nRxPktBusy);
		ASSERT(Adapter->nRxPktBusy >= 0);
		SMSC_TRACE1(DBG_RX,"RCB %p removed from RxPktBusy list and inserted into RxPktFreeList\r\n", pRCB);
		InsertTailList(&Adapter->RxPktFreeList, &pRCB->List);
		pRCB->ulSize = 0;

		NdisReleaseSpinLock(&Adapter->RxPktLock);

		if (!MP_TEST_FLAG(Adapter, fMP_HALT_IN_PROGRESS) && !MP_TEST_FLAG(Adapter, fMP_RESET_IN_PROGRESS))
		{
			if (!IsListEmpty(&Adapter->BulkInWaitList))
			{
				// We have RCBs waiting for packets and a packet just got freed,
				// process the RCB then, but do it in the workitem to avoid recursion
				// if the packet was returned by NDIS synchronously on indication and
				// we are running in the context of a completion routine already.

				if (InterlockedExchange(&Adapter->IsBulkInResumeWorkItemQueued, TRUE) == FALSE)
				{
					Adapter->nBulkInResumeWorkItemScheduled++;
					MP_INC_REF(Adapter);
					NdisScheduleWorkItem(&Adapter->BulkInResumeWorkItem);
				}
			}
		}
	}

	SMSC_TRACE1(DBG_RX,"<-- NICFreeRCB. nBusyRecv = %d\r\n", Adapter->nBusyRecv);
}


/*++

Routine Description:

	Check if the destination address of a received packet
	matches the receive criteria of our adapter.

Arguments:

	Adapter    - pointer to the adapter structure
	pDstMac - Destination MAC address to compare

Return Value:

	True or False

--*/
BOOLEAN
NICIsPacketAcceptable(
	IN PMP_ADAPTER Adapter,
	IN PUCHAR	pDstMac
	)
{
	UINT			AddrCompareResult;
	ULONG			PacketFilter;
	BOOLEAN 		bPacketMatch;
	BOOLEAN 		bIsMulticast, bIsBroadcast;
	SMSC_TRACE0(DBG_RX,"-->NICIsPacketAcceptable");
	PacketFilter = Adapter->PacketFilter;

	bIsMulticast = ETH_IS_MULTICAST(pDstMac);
	bIsBroadcast = ETH_IS_BROADCAST(pDstMac);

	//
	// Handle the directed packet case first.
	//
	if (!bIsMulticast)
	{
		//
		// If the Adapter is not in promisc. mode, check if
		// the destination MAC address matches the local
		// address.
		//
		if ((PacketFilter & NDIS_PACKET_TYPE_PROMISCUOUS) == 0)
		{
			ETH_COMPARE_NETWORK_ADDRESSES_EQ(Adapter->CurrentNodeAddress,
											 pDstMac,
											 &AddrCompareResult);

			bPacketMatch = ((AddrCompareResult == 0) &&
						   ((PacketFilter & NDIS_PACKET_TYPE_DIRECTED) != 0));
		}
		else
		{
			bPacketMatch = TRUE;
		}
	 }
	 else
	 {
		//
		// Multicast or broadcast packet.
		//

		//
		// Indicate if the filter is set to promisc mode ...
		//
		if ((PacketFilter & NDIS_PACKET_TYPE_PROMISCUOUS)
				||

			//
			// or if this is a broadcast packet and the filter
			// is set to receive all broadcast packets...
			//
			(bIsBroadcast &&
			 (PacketFilter & NDIS_PACKET_TYPE_BROADCAST))
				||

			//
			// or if this is a multicast packet, and the filter is
			// either set to receive all multicast packets, or
			// set to receive specific multicast packets. In the
			// latter case, indicate receive only if the destn
			// MAC address is present in the list of multicast
			// addresses set on the Adapter.
			//
			(!bIsBroadcast &&
			 ((PacketFilter & NDIS_PACKET_TYPE_ALL_MULTICAST) ||
			  ((PacketFilter & NDIS_PACKET_TYPE_MULTICAST) &&
			   NICIsMulticastMatch(Adapter, pDstMac))))
		   )
		{
			bPacketMatch = TRUE;
		}
		else
		{
			//
			// No protocols above are interested in this
			// multicast/broadcast packet.
			//
			bPacketMatch = FALSE;
		}
	}
SMSC_TRACE0(DBG_RX,"<--NICIsPacketAcceptable");
	return (bPacketMatch);
}


/*++

Routine Description:

	Check if the given multicast destination MAC address matches
	any of the multicast address entries set on the Adapter.

	NOTE: the caller is assumed to hold a READ/WRITE lock
	to the parent ADAPT structure. This is so that the multicast
	list on the Adapter is invariant for the duration of this call.

Arguments:

	Adapter  - Adapter to look in
	pDstMac - Destination MAC address to compare

Return Value:

	TRUE iff the address matches an entry in the Adapter

--*/
BOOLEAN
NICIsMulticastMatch(
	IN PMP_ADAPTER	Adapter,
	IN PUCHAR		pDstMac
	)
{
	ULONG			i;
	UINT			AddrCompareResult;
SMSC_TRACE0(DBG_MULTICAST,"-->NICIsMulticastMatch");
	for (i = 0; i < Adapter->ulMCListSize; i++)
	{
		ETH_COMPARE_NETWORK_ADDRESSES_EQ(Adapter->MCList[i],
										 (pDstMac+2),
										 &AddrCompareResult);

		if (AddrCompareResult == 0)
		{
			break;
		}
	}
SMSC_TRACE0(DBG_MULTICAST,"<--NICIsMulticastMatch");
	return (i != Adapter->ulMCListSize);
}


BOOLEAN
NICPacketPassesRxFilter(
    IN PMP_ADAPTER  Adapter,
    IN PDWORD       pData
    )
/*++

Routine Description:

    Check if the destination address of a received packet
    matches the receive criteria of our adapter.
    
Arguments:

    Adapter    - pointer to the adapter structure
    pData      - Pointer so the packet data read from the device, 
                 starting at it's Rx status word.

Return Value:

    True or False

--*/
{
    ULONG           PacketFilter;
    BOOLEAN         bFilterPass;

       PacketFilter = Adapter->PacketFilter;

    SMSC_TRACE1(DBG_RX,"RxSts = 0x%08lx\r\n", *pData);

    if ((((*pData) & RX_CMD_A_RED_) == 0) ||
        (((*pData) & RX_CMD_A_RX_ERRS_) == ((*pData) & RX_CMD_A_CSE_)))
    {
        //
        // Packet has no Receive Errors
        //
        if ((PacketFilter & NDIS_PACKET_TYPE_PROMISCUOUS) == 0)
        {
            //
            // We are not operating in promiscuous mode
            //

            if (((*pData) & (RX_CMD_A_MAM_ | RX_CMD_A_BAM_)) == 0)   
            {
                //
                // Packet is directed (NOT multicast / NOT broadcast). 
                //
    
                //
                // We'll pass it only if directed packet reception is enabled
                //
                // The hardware will filter the packets if they are not directed 
                // us so no need to check that.
                //
                SMSC_TRACE0(DBG_RX,"DIRECTED RxPkt\r\n");
                bFilterPass = (BOOLEAN) ((PacketFilter & NDIS_PACKET_TYPE_DIRECTED) != 0);
            }
            else
            {
                //
                // Multicast or broadcast packet.
                //
                if (((*pData) & RX_CMD_A_BAM_) == RX_CMD_A_BAM_)
                {
                    //
                    // Broadcast packet 
                    //
    
                    // The hardware will filter them if brodcast reception is 
                    // disabled so no need to check for that
                    SMSC_TRACE0(DBG_RX,"BROADCAST RxPkt\r\n");
                    bFilterPass = TRUE;
                }
                else
                {
                    //
                    // Multicast packet
                    // 
                    // Pass it only if one of these two cases is true:
                    // - we're set to receive all multicast packets
                    // - we're set to receive specific multicast packets
                    //   and the destination address is in the multicast 
                    //   list for the adapter.
                    SMSC_TRACE0(DBG_RX,"MULTICAST RxPkt\r\n");
                    if  (
                            (PacketFilter & NDIS_PACKET_TYPE_ALL_MULTICAST) ||
                            (
                                (PacketFilter & NDIS_PACKET_TYPE_MULTICAST) &&
                                 NICIsMulticastMatch(Adapter, (PUCHAR) &pData[2]) 
                            )
                        )
                    {
                        bFilterPass = TRUE;
                    }
                    else
                    {
                        bFilterPass = FALSE;
                    }
                }
            }
        }
        else
        {
            //
            // We are operating in promiscuous mode. Always indicate pkt.
            //
            bFilterPass = TRUE;
        }
    }
    else
    {
        SMSC_TRACE1(DBG_ERROR,"Rx ERROR: RxSts = 0x%08lx\n", *pData);
        if (Adapter->RxBadPkts)
            bFilterPass = TRUE;
        else
            bFilterPass = FALSE;
    }

    return (bFilterPass);
}


